@@ -11,6 +11,8 @@ import copy
 from collections.abc import Iterable
 import re
 
+from ..core.quant_analysis import ParamQuantitativeAnalysis
+
 from skbeam.core.fitting.xrf_model import (linear_spectrum_fitting, define_range)
 from .fileio import output_data, read_hdf_APS, read_MAPS, sep_v
 from .fit_spectrum import (single_pixel_fitting_controller,
@@ -24,6 +26,9 @@ def fit_pixel_data_and_save(working_directory, file_name, *,
                             param_file_name, fit_channel_sum=True,
                             fit_channel_each=False, param_channel_list=None,
                             incident_energy=None,
+                            ignore_datafile_metadata=False,
+                            fln_quant_calib_data=None,
+                            quant_distance_to_sample=0,
                             method='nnls', pixel_bin=0, raise_bg=0,
                             comp_elastic_combine=False,
                             linear_bg=False,
@@ -32,8 +37,9 @@ def fit_pixel_data_and_save(working_directory, file_name, *,
                             spectrum_cut=3000,
                             save_txt=False,
                             save_tiff=True,
-                            ic_name=None,
-                            use_average=True,
+                            scaler_name=None,
+                            use_average=False,
+                            interpolate_to_uniform_grid=False,
                             data_from='NSLS-II'):
     """
     Do fitting for signle data set, and save data accordingly. Fitting can be performed on
@@ -55,6 +61,19 @@ def fit_pixel_data_and_save(working_directory, file_name, *,
         list of param file names for each channel
     incident_energy : float, optional
         use this energy as incident energy instead of the one in param file, i.e., XANES
+        This value overrides the incident energy from metadata and from JSON parameter file
+    ignore_datafile_metadata : bool
+        tells whether to ignore metadata from the data file (if data file contains metadata).
+        At the moment, only incident energy (monochromator energy) is used by the processing routine.
+        If True, then the incident energy from the parameter JSON file is used; if False, then
+        the energy from metadata is used. If the function parameter ``incident_energy`` is
+        not None, then its value overrides the incident energy from metadata or JSON file.
+    fln_quant_calib_data : str or list(str)
+        file name or a list of file names that contain quantitative calibration data
+    quant_distance_to_sample : float
+        distance-to-sample used in quantitative calibration. If 0, then correction for
+        distance is not applied (assumed that the standard and the sample were placed
+        at the same distance from the detector).
     method : str, optional
         fitting method, default as nnls
     pixel_bin : int, optional
@@ -75,26 +94,51 @@ def fit_pixel_data_and_save(working_directory, file_name, *,
         save data to txt or not
     save_tiff : bool, optional
         save data to tiff or not
-    ic_name : str, optional
-        if given, normalization will be performed before saving data as .txt and .tiff
+    scaler_name : str, optional
+        name of the field representing the scaler (for example 'i0'), which must be present
+        in the data files. If given, normalization will be performed before saving data
+        as .txt and .tiff
     use_average : bool, optional
-        if true, norm is performed as data/IC*mean(IC), otherwise just data/IC
+        if true, norm is performed as data/scaler*mean(scaler), otherwise just data/scaler
+    interpolate_to_uniform_grid : bool
+        interpolate the result to uniform grid before saving to tiff and txt files
+        The grid dimensions match the dimensions of positional data for X and Y axes.
+        The range of axes is chosen to fit the values of X and Y.
     data_from : str, optional
         where do data come from? Data format includes data from NSLS-II, or 2IDE-APS
     """
     fpath = os.path.join(working_directory, file_name)
+
+    # Load quantitative calibration files (if necessary)
+    quant_norm = False  # Indicates if at least one calibration file is loaded
+    param_quant_analysis = ParamQuantitativeAnalysis()
+    if fln_quant_calib_data:
+        if isinstance(fln_quant_calib_data, str):
+            fln_quant_calib_data = [fln_quant_calib_data]
+        for fln in fln_quant_calib_data:
+            if os.path.isabs(fln):
+                f = fln
+            else:
+                f = os.path.join(working_directory, fln)
+            try:
+                param_quant_analysis.load_entry(f)
+                quant_norm = True
+                logger.info(f"Quantitative calibration is loaded successfully from file '{f}'")
+            except Exception as ex:
+                logger.error(f"Error occurred while loading quantitative calibration from file '{f}': {ex}")
+
     t0 = time.time()
     prefix_fname = file_name.split('.')[0]
     if fit_channel_sum is True:
         if data_from == 'NSLS-II':
-            img_dict, data_sets = read_hdf_APS(working_directory, file_name,
-                                               spectrum_cut=spectrum_cut,
-                                               load_each_channel=False)
+            img_dict, data_sets, mdata = read_hdf_APS(working_directory, file_name,
+                                                      spectrum_cut=spectrum_cut,
+                                                      load_each_channel=False)
         elif data_from == '2IDE-APS':
-            img_dict, data_sets = read_MAPS(working_directory,
-                                            file_name, channel_num=1)
+            img_dict, data_sets, mdata = read_MAPS(working_directory,
+                                                   file_name, channel_num=1)
         else:
-            print('Unkonw data sets.')
+            print('Unknown data sets.')
 
         try:
             data_all_sum = data_sets[prefix_fname+'_sum'].raw_data
@@ -112,6 +156,17 @@ def fit_pixel_data_and_save(working_directory, file_name, *,
         # update incident energy, required for XANES
         if incident_energy is not None:
             param_sum['coherent_sct_energy']['value'] = incident_energy
+            print("Using incident beam energy passed as the function parameter.")
+        elif mdata.is_metadata_available() and "instrument_mono_incident_energy" in mdata \
+                and not ignore_datafile_metadata:
+            param_sum['coherent_sct_energy']['value'] = mdata["instrument_mono_incident_energy"]
+            print(f"Using incident beam energy from the data file '{file_name}'.")
+        else:
+            print(f"Using incident beam energy from the parameter file '{param_path}'.")
+
+        # The value of incident energy that is used for processing
+        incident_energy_used = param_sum['coherent_sct_energy']['value']
+        print(f"Incident beam energy: {incident_energy_used}.")
 
         result_map_sum, calculation_info = single_pixel_fitting_controller(
             data_all_sum,
@@ -130,29 +185,112 @@ def fit_pixel_data_and_save(working_directory, file_name, *,
         # fit_name = prefix_fname+'_fit'
         save_fitdata_to_hdf(fpath, result_map_sum, datapath=inner_path)
 
-    if fit_channel_each is True and param_channel_list is not None:
+        def get_scaler_set(img_dict):
+            sc_set_names = [_ for _ in img_dict if _.endswith("_scaler")]
+            if sc_set_names:
+                return img_dict[sc_set_names[0]]
+            else:
+                return {}
+
+        def get_positions_set(img_dict):
+            if "positions" in img_dict:
+                return img_dict["positions"]
+            else:
+                return {}
+
+        scaler_dict = get_scaler_set(img_dict)
+        scaler_name_list = list(scaler_dict.keys())
+        positions_dict = get_positions_set(img_dict)
+        # Generate dataset
+        dataset = copy.deepcopy(scaler_dict)
+        dataset.update(result_map_sum)
+
+        # Set parameters for quantitative normalization
+        param_quant_analysis.experiment_incident_energy = incident_energy_used
+        param_quant_analysis.experiment_distance_to_sample = quant_distance_to_sample
+        param_quant_analysis.experiment_detector_channel = "sum"
+
+        if save_txt is True:
+            output_folder = 'output_txt_'+prefix_fname
+            output_path = os.path.join(working_directory, output_folder)
+            output_data(output_dir=output_path,
+                        interpolate_to_uniform_grid=interpolate_to_uniform_grid,
+                        dataset_name="dataset_fit",  # Sum of all detectors: should end with '_fit'
+                        quant_norm=quant_norm,
+                        param_quant_analysis=param_quant_analysis,
+                        distance_to_sample=quant_distance_to_sample,
+                        dataset_dict=dataset, positions_dict=positions_dict,
+                        file_format="txt", scaler_name=scaler_name,
+                        scaler_name_list=scaler_name_list,
+                        use_average=use_average)
+
+        if save_tiff is True:
+            output_folder = 'output_tiff_'+prefix_fname
+            output_path = os.path.join(working_directory, output_folder)
+            output_data(output_dir=output_path,
+                        interpolate_to_uniform_grid=interpolate_to_uniform_grid,
+                        dataset_name="dataset_fit",  # Sum of all detectors: should end with '_fit'
+                        quant_norm=quant_norm,
+                        param_quant_analysis=param_quant_analysis,
+                        dataset_dict=dataset, positions_dict=positions_dict,
+                        file_format="tiff", scaler_name=scaler_name,
+                        scaler_name_list=scaler_name_list,
+                        use_average=use_average)
+
+    if fit_channel_each:
+        img_dict, data_sets, mdata = read_hdf_APS(working_directory, file_name,
+                                                  spectrum_cut=spectrum_cut,
+                                                  load_each_channel=True)
+
+        # Find the detector channels and the names of the channels
+        det_channels = [_ for _ in data_sets.keys() if re.search(r"_det\d+$", _)]
+        det_channel_names = [re.search(r"det\d+$", _)[0] for _ in det_channels]
+        if param_channel_list is None:
+            param_channel_list = [param_file_name] * 3
+        else:
+            if not isinstance(param_channel_list, list) and not isinstance(param_channel_list, tuple):
+                raise RuntimeError(f"Parameter 'param_channel_list' must be a list or a tuple of strings")
+            if len(param_channel_list) != len(det_channels):
+                raise RuntimeError(f"Parameter 'param_channel_list' must be 'None' "
+                                   f"or contain {len(det_channels)} file names.")
+
         channel_num = len(param_channel_list)
-        img_dict, data_sets = read_hdf_APS(working_directory, file_name,
-                                           spectrum_cut=spectrum_cut,
-                                           load_each_channel=True)
         for i in range(channel_num):
-            filename_det = prefix_fname+'_det'+str(i+1)
-            inner_path = 'xrfmap/det'+str(i+1)
+            inner_path = 'xrfmap/' + det_channel_names[i]
+            print(f"Processing data from detector channel {det_channel_names[i]} (#{i+1}) ...")
 
             # load param file
-            param_file_det = param_channel_list[i]
-            param_path = os.path.join(working_directory, param_file_det)
+            param_file_name = param_channel_list[i]
+
+            # load param file
+            if not os.path.isabs(param_file_name):
+                param_path = os.path.join(working_directory, param_file_name)
+            else:
+                param_path = param_file_name
             with open(param_path, 'r') as json_data:
                 param_det = json.load(json_data)
 
             # update incident energy, required for XANES
             if incident_energy is not None:
                 param_det['coherent_sct_energy']['value'] = incident_energy
+                print(f"Using incident beam energy passed as the function parameter.")
+            elif mdata.is_metadata_available() and "instrument_mono_incident_energy" in mdata \
+                    and not ignore_datafile_metadata:
+                param_det['coherent_sct_energy']['value'] = mdata["instrument_mono_incident_energy"]
+                print(f"Using incident beam energy from the data file '{file_name}'.")
+            else:
+                print(f"Using incident beam energy from the parameter file '{param_path}'.")
+
+            # The value of incident energy that is used for processing
+            incident_energy_used = param_det['coherent_sct_energy']['value']
+            print(f"Incident beam energy: {incident_energy_used}.")
+
+            data_all_det = data_sets[det_channels[i]].raw_data
 
-            data_all_det = data_sets[filename_det].raw_data
             result_map_det, calculation_info = single_pixel_fitting_controller(
                 data_all_det,
                 param_det,
+                incident_energy=incident_energy,
                 method=method,
                 pixel_bin=pixel_bin,
                 raise_bg=raise_bg,
@@ -160,25 +298,72 @@ def fit_pixel_data_and_save(working_directory, file_name, *,
                 linear_bg=linear_bg,
                 use_snip=use_snip,
                 bin_energy=bin_energy)
+
             # output to .h5 file
             save_fitdata_to_hdf(fpath, result_map_det, datapath=inner_path)
 
+            def get_scaler_set(img_dict):
+                sc_set_names = [_ for _ in img_dict if _.endswith("_scaler")]
+                if sc_set_names:
+                    return img_dict[sc_set_names[0]]
+                else:
+                    return {}
+
+            def get_positions_set(img_dict):
+                if "positions" in img_dict:
+                    return img_dict["positions"]
+                else:
+                    return {}
+
+            scaler_dict = get_scaler_set(img_dict)
+            scaler_name_list = list(scaler_dict.keys())
+            positions_dict = get_positions_set(img_dict)
+            # Generate dataset
+            dataset = copy.deepcopy(scaler_dict)
+            dataset.update(result_map_det)
+
+            # Set parameters for quantitative normalization
+            param_quant_analysis.experiment_incident_energy = incident_energy_used
+            param_quant_analysis.experiment_distance_to_sample = quant_distance_to_sample
+            param_quant_analysis.experiment_detector_channel = det_channel_names[i]
+
+            if save_txt is True:
+                output_folder = 'output_txt_'+prefix_fname
+                output_path = os.path.join(working_directory, output_folder)
+                output_data(output_dir=output_path,
+                            interpolate_to_uniform_grid=interpolate_to_uniform_grid,
+                            dataset_name=f"dataset_{det_channel_names[i]}_fit",  # ..._det1_fit, etc.
+                            quant_norm=quant_norm,
+                            param_quant_analysis=param_quant_analysis,
+                            dataset_dict=dataset, positions_dict=positions_dict,
+                            file_format="txt", scaler_name=scaler_name,
+                            scaler_name_list=scaler_name_list,
+                            use_average=use_average)
+
+            if save_tiff is True:
+                output_folder = 'output_tiff_'+prefix_fname
+                output_path = os.path.join(working_directory, output_folder)
+                output_data(output_dir=output_path,
+                            interpolate_to_uniform_grid=interpolate_to_uniform_grid,
+                            dataset_name=f"dataset_{det_channel_names[i]}_fit",  # ..._det1_fit, etc.
+                            quant_norm=quant_norm,
+                            param_quant_analysis=param_quant_analysis,
+                            dataset_dict=dataset, positions_dict=positions_dict,
+                            file_format="tiff", scaler_name=scaler_name,
+                            scaler_name_list=scaler_name_list,
+                            use_average=use_average)
+
     t1 = time.time()
-    print(f"Time used for pixel fitting for file '{file_name}' is : {t1 - t0}")
-
-    if save_txt is True:
-        output_folder = 'output_txt_'+prefix_fname
-        output_path = os.path.join(working_directory, output_folder)
-        output_data(fpath, output_path, file_format='txt', norm_name=ic_name, use_average=use_average)
-    if save_tiff is True:
-        output_folder = 'output_tiff_'+prefix_fname
-        output_path = os.path.join(working_directory, output_folder)
-        output_data(fpath, output_path, file_format='tiff', norm_name=ic_name, use_average=use_average)
+    print(f"Processing time: {t1 - t0}")
 
 
-def pyxrf_batch(start_id=None, end_id=None, *, param_file_name, data_files=None, wd=None, fit_channel_sum=True,
-                fit_channel_each=False, param_channel_list=None, incident_energy=None,
-                spectrum_cut=3000, save_txt=False, save_tiff=True, ic_name=None, use_average=True):
+def pyxrf_batch(start_id=None, end_id=None, *, param_file_name, data_files=None, wd=None,
+                fit_channel_sum=True, fit_channel_each=False, param_channel_list=None,
+                incident_energy=None, ignore_datafile_metadata=False,
+                fln_quant_calib_data=None, quant_distance_to_sample=0,
+                spectrum_cut=3000, save_txt=False, save_tiff=True,
+                scaler_name=None, use_average=False,
+                interpolate_to_uniform_grid=False):
     """
     Perform fitting on a batch of data files. The results are saved as new datasets
     in the respective data files and may be viewed using PyXRF. Fitting can be performed on
@@ -213,16 +398,34 @@ def pyxrf_batch(start_id=None, end_id=None, *, param_file_name, data_files=None,
         list of param file names for each channel
     incident_energy : float, optional
         use this energy as incident energy instead of the one in param file, i.e., XANES
+        This value overrides the incident energy from metadata and from JSON parameter file.
+    ignore_datafile_metadata : bool
+        tells whether to ignore metadata from the data file (if data file contains metadata).
+        At the moment, only incident energy (monochromator energy) is used by the processing routine.
+        If True, then the incident energy from the parameter JSON file is used; if False, then
+        the energy from metadata is used. If the function parameter ``incident_energy`` is
+        not None, then its value overrides the incident energy from metadata or JSON file.
+        Default: False.
+    fln_quant_calib_data : str or list(str)
+        file name or a list of file names that contain quantitative calibration data
+    quant_distance_to_sample : float
+        distance-to-sample used in quantitative calibration. If 0, then correction for
+        distance is not applied (assumed that the standard and the sample were placed
+        at the same distance from the detector).
     spectrum_cut : int, optional
         only use spectrum from, say 0, 3000
     save_txt : bool, optional
         save data to txt or not
     save_tiff : bool, optional
         save data to tiff or not
-    ic_name : str, optional
+    scaler_name : str, optional
         if given, normalization will be performed
     use_average : bool, optional
-        if true, norm is performed as data/IC*mean(IC), otherwise just data/IC
+        if true, norm is performed as data/scaler*mean(scaler), otherwise just data/scaler
+    interpolate_to_uniform_grid : bool
+        interpolate the result to uniform grid before saving to tiff and txt files
+        The grid dimensions match the dimensions of positional data for X and Y axes.
+        The range of axes is chosen to fit the values of X and Y.
 
     Returns
     -------
@@ -322,6 +525,16 @@ def pyxrf_batch(start_id=None, end_id=None, *, param_file_name, data_files=None,
     elif (start_id is not None) and (end_id is None):
         allow_raising_exceptions = True
 
+    if fln_quant_calib_data is not None:
+        if not isinstance(fln_quant_calib_data, str) and not isinstance(fln_quant_calib_data, list):
+            raise ValueError("Parameter 'fln_quant_calib_data' must be a string or a list of strings")
+        if isinstance(fln_quant_calib_data, list) and \
+                any([not isinstance(_, str) for _ in fln_quant_calib_data]):
+            raise ValueError("List passed with the parameter 'fln_quant_calib_data' must contain only strings")
+
+    if not isinstance(quant_distance_to_sample, float) and not isinstance(quant_distance_to_sample, int):
+        raise ValueError("Value of the parameter 'quant_distance_to_sample' must be floating point number")
+
     if data_files is not None:
 
         # Check if ``data_files`` has valid value
@@ -354,9 +567,9 @@ def pyxrf_batch(start_id=None, end_id=None, *, param_file_name, data_files=None,
                 flist.append(fln)
             else:
                 if allow_raising_exceptions:
-                    raise IOError(f"File {fln} does not exist")
+                    raise IOError(f"File '{fln}' does not exist")
                 else:
-                    print(f"WARNING: file {fln} does not exist.")
+                    print(f"WARNING: file '{fln}' does not exist.")
 
     else:
 
@@ -364,6 +577,9 @@ def pyxrf_batch(start_id=None, end_id=None, *, param_file_name, data_files=None,
 
         all_files = glob.glob(os.path.join(wd, '*.h5'))
 
+        # Sort files (processing unsorted list of files is unsightly)
+        all_files.sort()
+
         if start_id is None and end_id is None:
             # ``start_id`` and ``end_id`` are not specified:
             #   process all .h5 files in the current directory
@@ -374,6 +590,7 @@ def pyxrf_batch(start_id=None, end_id=None, *, param_file_name, data_files=None,
             #   (only if such file exists)
             flist = [fname for fname in all_files
                      if re.search(f"^[^_]*_{str(start_id)}\D+", os.path.basename(fname))]  # noqa: W605
+
             if len(flist) < 1:
                 msg = f"File with Scan ID {start_id} was not found"
                 if allow_raising_exceptions:
@@ -402,7 +619,7 @@ def pyxrf_batch(start_id=None, end_id=None, *, param_file_name, data_files=None,
     if not os.path.exists(pname) or not os.path.isfile(pname):
         raise IOError(f"Function 'pyxrf_batch'. Processing parameter file '{pname}' does not exist.")
     else:
-        print(f"Processing parameter file: {pname}")
+        print(f"Processing parameter file: '{pname}'")
 
     if len(flist) > 0:
 
@@ -419,9 +636,14 @@ def pyxrf_batch(start_id=None, end_id=None, *, param_file_name, data_files=None,
                 fit_pixel_data_and_save(working_directory, fname,
                                         fit_channel_sum=fit_channel_sum, param_file_name=param_file_name,
                                         fit_channel_each=fit_channel_each, param_channel_list=param_channel_list,
-                                        incident_energy=incident_energy, spectrum_cut=spectrum_cut,
+                                        incident_energy=incident_energy,
+                                        ignore_datafile_metadata=ignore_datafile_metadata,
+                                        fln_quant_calib_data=fln_quant_calib_data,
+                                        quant_distance_to_sample=quant_distance_to_sample,
+                                        spectrum_cut=spectrum_cut,
                                         save_txt=save_txt, save_tiff=save_tiff,
-                                        ic_name=ic_name, use_average=use_average)
+                                        scaler_name=scaler_name, use_average=use_average,
+                                        interpolate_to_uniform_grid=interpolate_to_uniform_grid)
             except Exception as ex:
                 if allow_raising_exceptions:
                     raise Exception from ex
@@ -569,3 +791,34 @@ def fit_data_multi_files(dir_path, file_prefix,
     pool.terminate()
     pool.join()
     return results
+
+
+if __name__ == "__main__":
+    logger = logging.getLogger()
+
+    logger.setLevel(logging.INFO)
+
+    formatter = logging.Formatter(fmt='%(asctime)s : %(levelname)s : %(message)s')
+
+    stream_handler = logging.StreamHandler()
+    stream_handler.setFormatter(formatter)
+    stream_handler.setLevel(logging.INFO)
+    logger.addHandler(stream_handler)
+
+    # pyxrf_batch(start_id=92276, end_id=92276,
+    #             param_file_name="param_335",
+    #             wd=".", save_tiff=False)
+
+    pyxrf_batch(start_id=63339,
+                param_file_name="test",
+                wd=".", save_tiff=True,
+                fln_quant_calib_data=['standard_41147a.json'],
+                quant_distance_to_sample=2.0,
+                scaler_name="i0",
+                # fln_quant_calib_data=['standard_32654A.json',
+                #                       'standard_32654A a.json',
+                #                       'standard_41147.json'],
+                fit_channel_sum=True,
+                fit_channel_each=True,
+                # param_channel_list=['det1', 'det2', 'det3']
+                )
