@@ -6,6 +6,7 @@ import sys
 import h5py
 import numpy as np
 import os
+import re
 from collections import OrderedDict
 import pandas as pd
 import json
@@ -16,10 +17,12 @@ import glob
 import ast
 import matplotlib.animation as animation
 import matplotlib.pyplot as plt
-from atom.api import Atom, Str, observe, Typed, Dict, List, Int, Enum, Bool
+from atom.api import Atom, Str, observe, Typed, Dict, List, Int, Float, Enum, Bool
 from .load_data_from_db import (db, fetch_data_from_db, flip_data,
                                 helper_encode_list, helper_decode_list,
                                 write_db_to_hdf)
+from ..core.utils import normalize_data_by_scaler, grid_interpolate
+from .scan_metadata import ScanMetadataXRF
 import requests
 from distutils.version import LooseVersion
 
@@ -95,6 +98,19 @@ class FileIOModel(Atom):
 
     data_ready = Bool(False)
 
+    # Scan metadata
+    scan_metadata = Typed(ScanMetadataXRF)
+    # Indicates if metadata is available for recently loaded scan
+    scan_metadata_available = Bool(False)
+    # Indicates if the incident energy is available in metadata for recently loaded scan
+    incident_energy_available = Bool(False)
+
+    # Changing this variable sets incident energy in the ``plot_model``
+    #   Must be linked with the function ``plot_model.set_incident_energy``
+    # This value is not updated if incident energy parameter is changed somewhere else, therefore
+    #   its value should not be used for computations!!!
+    incident_energy_set = Float(0.0)
+
     def __init__(self, **kwargs):
         self.working_directory = kwargs['working_directory']
         self.mask_data = None
@@ -170,6 +186,34 @@ class FileIOModel(Atom):
     def window_title_set_run_id(self, run_id):
         self.window_title = f"{self.window_title_base} - Scan ID: {run_id}"
 
+    def _metadata_update_program_state(self):
+        """
+        Update program state based on metadata:
+
+        -- enable controls (mostly ``View Metadata`` button in ``File IO`` tab
+
+        -- set incident energy if it is available
+
+        -- print logger warning if incident energy is not available in metadata
+        (or if metadata does not exist)
+        """
+
+        self.scan_metadata_available = False
+        self.incident_energy_available = False
+        if self.scan_metadata is not None:
+            self.scan_metadata_available = self.scan_metadata.is_metadata_available()
+            self.incident_energy_available = self.scan_metadata.is_mono_incident_energy_available()
+
+        if self.incident_energy_available:
+            # Fetch incident energy from metadata if it exists
+            self.incident_energy_set = self.scan_metadata.get_mono_incident_energy()
+            logger.info(f"Incident energy {self.incident_energy_set} keV was extracted from the scan metadata")
+        else:
+            logger.warning(
+                "Incident energy is not available in scan metadata and needs to be set manually: "
+                "click 'Find Elements Automatically' button in 'Fit' "
+                "tab to access the settings dialog box.")
+
     @observe(str('file_name'))
     def update_more_data(self, change):
         if change['value'] == 'temp':
@@ -185,9 +229,13 @@ class FileIOModel(Atom):
         logger.info('File is loaded: %s' % (self.file_name))
 
         # focus on single file only
-        self.img_dict, self.data_sets = file_handler(self.working_directory,
-                                                     self.file_name,
-                                                     load_each_channel=self.load_each_channel)
+        self.img_dict, self.data_sets, self.scan_metadata = \
+            file_handler(self.working_directory,
+                         self.file_name,
+                         load_each_channel=self.load_each_channel)
+
+        # Process metadata
+        self._metadata_update_program_state()
 
         self.data_ready = True
 
@@ -222,13 +270,23 @@ class FileIOModel(Atom):
         #                                             self.fname_from_db,
         #                                             load_each_channel=self.load_each_channel)
 
-        rv = render_data_to_gui(self.runid, file_overwrite_existing=self.file_overwrite_existing)
+        rv = render_data_to_gui(self.runid,
+                                create_each_det=self.load_each_channel,
+                                working_directory=self.working_directory,
+                                file_overwrite_existing=self.file_overwrite_existing)
 
         if rv is None:
             logger.error(f"Data from scan #{self.runid} was not loaded")
             return
 
-        img_dict, self.data_sets, fname, detector_name = rv
+        img_dict, self.data_sets, fname, detector_name, self.scan_metadata = rv
+
+        # Replace relative scan ID with true scan ID.
+        if (self.runid) < 0 and ("scan_id" in self.scan_metadata):
+            self.runid = int(self.scan_metadata["scan_id"])
+
+        # Process metadata
+        self._metadata_update_program_state()
 
         # Change file name without rereading the file
         self.file_name_silent_change = True
@@ -289,6 +347,21 @@ class FileIOModel(Atom):
         # get summed data or based on mask
         self.data = self.data_sets[self.selected_file_name].get_sum()
 
+    def get_selected_detector_channel(self):
+        r"""
+        Returns selected channel name. Expected values are ``sum``, ``det1``, ``det2``, etc.
+        If no channel is selected or it is impossible to determine the channel name, then
+        the return value is ``None`` (this is not a normal outcome).
+        """
+        det_channel = None
+        if self.selected_file_name:
+            try:
+                # The channel is supposed to be the last component of the 'selected_file_name'
+                det_channel = self.selected_file_name.split("_")[-1]
+            except Exception:
+                pass
+        return det_channel
+
     def apply_mask(self):
         """Apply mask with different options.
         """
@@ -383,12 +456,14 @@ class DataSelection(Atom):
     def get_sum(self, mask=None):
         if len(self.point1) == 0 and len(self.point2) == 0:
             SC = SpectrumCalculator(self.raw_data)
-            return SC.get_spectrum(mask=mask)
+            spec = SC.get_spectrum(mask=mask)
         else:
             SC = SpectrumCalculator(self.raw_data,
                                     pos1=self.point1,
                                     pos2=self.point2)
-            return SC.get_spectrum()
+            spec = SC.get_spectrum()
+        # Return the 'sum' spectrum as regular 64-bit float (raw data is in 'np.float32')
+        return spec.astype(np.float64, copy=False)
 
 
 class SpectrumCalculator(object):
@@ -496,82 +571,115 @@ def read_xspress3_data(file_path):
     return data_output
 
 
-def output_data(fpath, output_folder,
-                file_format='tiff', norm_name=None, use_average=True):
+def output_data(dataset_dict=None, output_dir=None,
+                file_format='tiff', scaler_name=None, use_average=False,
+                dataset_name=None, quant_norm=False,
+                param_quant_analysis=None,
+                positions_dict=None,
+                interpolate_to_uniform_grid=False,
+                scaler_name_list=None):
     """
     Read data from h5 file and transfer them into txt.
 
     Parameters
     ----------
-    fpath : str
-        path to h5 file
-    output_folder : str
+    dataset_dict : dict(ndarray)
+        Dictionary of XRF maps contained in the selected dataset. Each XRF map is saved in
+        the individual file. File name consists of the detector channel name (e.g. 'detsum', 'det1' etc.)
+        and map name (dictionary key). Optional list of scaler names 'scaler_name_list' may be passed
+        to the function. If map name is contained in the scaler name list, then the detector channel
+        name is not attached to the file name.
+    output_dir : str
         which folder to save those txt file
     file_format : str, optional
         tiff or txt
-    norm_name : str, optional
+    scaler_name : str, optional
         if given, normalization will be performed.
     use_average : Bool, optional
-        when normalization, multiply mean value of denomenator,
-        i.e., norm_data = data1/data2 * np.mean(data2)
+        when normalization, multiply by the mean value of scaler,
+        i.e., norm_data = data/scaler * np.mean(scaler)
+    dataset_name : str
+        the name of the selected datset (in Element Map tab)
+        should end with the suffix '_fit' (for sum of all channels), '_det1_fit" etc.
+    quant_norm : bool
+        True - quantitative normalization is enabled, False - disabled
+    param_quant_analysis : ParamQuantitativeAnalysis
+        reference to class, which contains parameters for quantitative normalization
+    interpolate_to_uniform_grid : bool
+        interpolate the result to uniform grid before saving to tiff and txt files
+        The grid dimensions match the dimensions of positional data for X and Y axes.
+        The range of axes is chosen to fit the values of X and Y.
+    scaler_name_list : list(str)
+        The list of names of scalers that may exist in the dataset 'dataset_dict'
     """
 
-    with h5py.File(fpath, 'r') as f:
-        tmp = output_folder.split(sep_v)[-1]
-        name_append = tmp.split('_')[-1]
-        if not name_append.isdigit():
-            name_append = ''
-        detlist = list(f['xrfmap'].keys())
-        fit_output = {}
+    if not dataset_name:
+        raise RuntimeError("Dataset is not selected. Data can not be saved.")
 
-        for detname in detlist:
-            # fitted data
-            if 'xrf_fit' in f['xrfmap/'+detname]:
-                fit_data = f['xrfmap/'+detname+'/xrf_fit']
-                fit_name = f['xrfmap/'+detname+'/xrf_fit_name']
-                fit_name = helper_decode_list(fit_name)
-                for i in np.arange(len(fit_name)):
-                    fit_output[detname+'_'+fit_name[i]] = np.asarray(fit_data[i, :, :])
-            # fitted error
-            if 'xrf_fit_error' in f['xrfmap/'+detname]:
-                error_data = f['xrfmap/'+detname+'/xrf_fit_error']
-                error_name = f['xrfmap/'+detname+'/xrf_fit_error_name']
-                error_name = helper_decode_list(error_name)
+    if dataset_dict is None:
+        dataset_dict = {}
 
-                for i in np.arange(len(error_name)):
-                    fit_output[detname+'_'+error_name[i]+'_error'] = np.asarray(error_data[i, :, :])
+    if positions_dict is None:
+        positions_dict = {}
 
-        # ic data
-        if 'scalers' in f['xrfmap']:
-            ic_data = f['xrfmap/scalers/val']
-            ic_name = f['xrfmap/scalers/name']
-            ic_name = helper_decode_list(ic_name)
-            for i in np.arange(len(ic_name)):
-                fit_output[ic_name[i]] = np.asarray(ic_data[:, :, i])
+    # Extract the detector channel name from dataset name
+    #   Dataset name ends with '_fit' for the sum of all channels
+    #   and '_detX_fit' for detector X (X is 1, 2, 3 ..)
+    # The extracted detector channel name should be 'detsum', 'det1', 'det2' etc.
+    dset = None
+    if re.search(r"_det\d+_fit$", dataset_name):
+        dset = re.search(r"_det\d_", dataset_name)[0]
+        dset = dset.strip('_')
+    elif re.search(r"_fit$", dataset_name):
+        dset = "detsum"
+    if not dset:
+        raise RuntimeError(f"Dataset '{dataset_name}' contains no useful data. "
+                           "Select different dataset to save data.")
 
-        # position data
-        if 'positions' in f['xrfmap']:
-            pos_name = f['xrfmap/positions/name']
-            pos_name = helper_decode_list(pos_name)
-            for i, n in enumerate(pos_name):
-                fit_output[n] = np.asarray(f['xrfmap/positions/pos'].value[i, :])
+    file_format = file_format.lower()
 
-    #  more data from suitcase part
-    data_sc = {}
-    # data_sc = retrieve_data_from_hdf_suitcase(fpath)
-    if len(data_sc) != 0:
-        fit_output.update(data_sc)
+    fit_output = {}
+    for k, v in dataset_dict.items():
+        fit_output[k] = v
 
-    output_data_to_tiff(fit_output, output_folder=output_folder,
-                        file_format=file_format, name_append=name_append,
-                        norm_name=norm_name,
-                        use_average=use_average)
+    for k, v in positions_dict.items():
+        fit_output[k] = v
+
+    logger.info(f"Saving data as {file_format.upper()} files. Directory '{output_dir}'")
+    if scaler_name:
+        logger.info(f"Data is NORMALIZED before saving. Scaler: '{scaler_name}'")
+
+    if(interpolate_to_uniform_grid):
+        if ("x_pos" in fit_output) and ("y_pos" in fit_output):
+            logger.info(f"Data is INTERPOLATED to uniform grid.")
+            for k, v in fit_output.items():
+                # Do not interpolation positions
+                if 'pos' in k:
+                    continue
+
+                fit_output[k], xx, yy = grid_interpolate(v, fit_output["x_pos"], fit_output["y_pos"])
+
+            fit_output["x_pos"] = xx
+            fit_output["y_pos"] = yy
+        else:
+            logger.error(f"Positional data 'x_pos' and 'y_pos' is not found in the dataset.\n"
+                         "Iterpolation to uniform grid can not be performed. "
+                         "Data is saved without interpolation.")
+
+    output_data_to_tiff(fit_output, output_dir=output_dir,
+                        file_format=file_format, name_prefix_detector=dset, name_append="",
+                        scaler_name=scaler_name, quant_norm=quant_norm,
+                        param_quant_analysis=param_quant_analysis,
+                        use_average=use_average,
+                        scaler_name_list=scaler_name_list)
 
 
 def output_data_to_tiff(fit_output,
-                        output_folder="~/pyxrf_data_tmp/",
-                        file_format='tiff', name_append="",
-                        norm_name=None, use_average=True):
+                        output_dir=None,
+                        file_format='tiff', name_prefix_detector=None, name_append=None,
+                        scaler_name=None, scaler_name_list=None,
+                        quant_norm=False, param_quant_analysis=None,
+                        use_average=False):
     """
     Read data in memory and save them into tiff to txt.
 
@@ -579,61 +687,137 @@ def output_data_to_tiff(fit_output,
     ----------
     fit_output:
         dict of fitting data and scaler data
-    output_folder : str, optional
+    output_dir : str, optional
         which folder to save those txt file
     file_format : str, optional
         tiff or txt
+    name_prefix_detector : str
+        prefix appended to file name except for the files that contain positional data and scalers
     name_append: str, optional
         more information saved to output file name
-    norm_name : str, optional
+    scaler_name : str, optional
         if given, normalization will be performed.
+    scaler_name_list : list(str)
+        The list of names of scalers that may exist in the dataset 'dataset_dict'
+    quant_norm : bool
+        True - apply quantitative normalization, False - use normalization by scaler
+    param_quant_analysis : ParamQuantitativeAnalysis
+        reference to class, which contains parameters for quantitative normalization,
+        if None, then quantitative normalization will be skipped
     use_average : Bool, optional
-        when normalization, multiply mean value of denomenator,
-        i.e., norm_data = data1/data2 * np.mean(data2)
+        when normalization, multiply by the mean value of scaler,
+        i.e., norm_data = data/scaler * np.mean(scaler)
     """
-    # save data
-    if os.path.exists(output_folder) is False:
-        logger.warning("Output_folder {} is created".format(output_folder))
-        os.mkdir(output_folder)
 
-    if norm_name is not None:
-        ic_v = fit_output[str(norm_name)]
-        norm_sign = 'norm'
-        for k, v in six.iteritems(fit_output):
-            if 'pos' in k or 'r2' in k:
-                continue
-            ave = 1.0
-            if use_average is True:
-                ave = np.mean(ic_v)
-            v = v/ic_v * ave
-            _fname = "_".join([k, name_append, norm_sign])
-            if file_format == 'tiff':
-                fname = os.path.join(output_folder, _fname + '.tiff')
-                sio.imsave(fname, v.astype(np.float32))
-            elif file_format == 'txt':
-                fname = os.path.join(output_folder, _fname + '.txt')
-                np.savetxt(fname, v.astype(np.float32))
+    if output_dir is None:
+        raise ValueError("Output directory is not specified.")
 
-    for k, v in six.iteritems(fit_output):
-        _fname = "_".join([k, name_append])
-        if file_format == 'tiff':
-            fname = os.path.join(output_folder, _fname + '.tiff')
-            sio.imsave(fname, v.astype(np.float32))
-        elif file_format == 'txt':
-            fname = os.path.join(output_folder, _fname + '.txt')
-            np.savetxt(fname, v.astype(np.float32))
+    if name_append:
+        name_append = f"_{name_append}"
+    else:
+        # If 'name_append' is None, set it to "" so that it could be safely appended to a string
+        name_append = ""
+
+    file_format = file_format.lower()
+
+    allowed_formats = ('txt', 'tiff')
+    if file_format not in allowed_formats:
+        raise RuntimeError(f"The specified format '{file_format}' not in {allowed_formats}")
+
+    # Create the output directory if it does not exist
+    os.makedirs(output_dir, exist_ok=True)
+
+    def _save_data(data, *, output_dir, file_name,
+                   name_prefix_detector, name_append,
+                   file_format, scaler_name_list):
+
+        # The 'file_format' is specified as file extension
+        file_extension = file_format.lower()
+
+        # If data is scalar or position, then don't attach the prefix
+        fname = f"{name_prefix_detector}_{file_name}" \
+            if (file_name not in scaler_name_list) and ("pos" not in file_name) \
+            else file_name
+
+        fname = f"{fname}{name_append}.{file_extension}"
+        fname = os.path.join(output_dir, fname)
+        if file_format.lower() == 'tiff':
+            sio.imsave(fname, data.astype(np.float32))
+        elif file_format.lower() == 'txt':
+            np.savetxt(fname, data.astype(np.float32))
+        else:
+            raise ValueError(f"Function is called with invalid file format '{file_format}'.")
+
+    if quant_norm:
+        if param_quant_analysis:
+            for data_name, data in fit_output.items():
+                # Quantitative normalization
+                data_normalized, quant_norm_applied = param_quant_analysis.apply_quantitative_normalization(
+                    data_in=data,
+                    scaler_dict=fit_output,
+                    scaler_name_default=None,  # We don't want data to be scaled
+                    data_name=data_name,
+                    name_not_scalable=None)  # For simplicity, all saved maps are normalized
+                if quant_norm_applied:
+                    # Save data only if quantitative normalization was performed.
+                    _save_data(data_normalized, output_dir=output_dir,
+                               file_name=data_name,
+                               name_prefix_detector=name_prefix_detector,
+                               name_append=f"{name_append}_quantitative",
+                               file_format=file_format,
+                               scaler_name_list=scaler_name_list)
+        else:
+            logger.error("Quantitative analysis parameters are not provided. "
+                         f"Quantitative data is not saved in {file_format.upper()} format.")
+
+    # Normalize data if scaler is provided
+    if scaler_name is not None:
+        if scaler_name in fit_output:
+            scaler_data = fit_output[scaler_name]
+            for data_name, data in fit_output.items():
+                if 'pos' in data_name or 'r2' in data_name:
+                    continue
+                # Normalization of data
+                data_normalized = normalize_data_by_scaler(data, scaler_data)
+                if use_average is True:
+                    data_normalized *= np.mean(scaler_data)
+
+                _save_data(data_normalized, output_dir=output_dir,
+                           file_name=data_name,
+                           name_prefix_detector=name_prefix_detector,
+                           name_append=f"{name_append}_norm",
+                           file_format=file_format,
+                           scaler_name_list=scaler_name_list)
+        else:
+            logger.warning(f"The scaler '{scaler_name}' was not found. Data normalization "
+                           f"was not performed for {file_format.upper()} file.")
+
+    # Always save not normalized data
+    for data_name, data in fit_output.items():
+        _save_data(data, output_dir=output_dir,
+                   file_name=data_name,
+                   name_prefix_detector=name_prefix_detector,
+                   name_append=name_append,
+                   file_format=file_format,
+                   scaler_name_list=scaler_name_list)
 
 
 def read_hdf_APS(working_directory,
                  file_name, spectrum_cut=3000,
-                 load_summed_data=True,
-                 load_each_channel=True):
+                 # The following parameters allow fine grained control over what is loaded from the file
+                 load_summed_data=True,  # Enable loading of RAW, FIT or ROI data from 'sum' channel
+                 load_each_channel=False,  # .. RAW data from individual detector channels
+                 load_processed_each_channel=True,  # .. FIT or ROI data from the detector channels
+                 load_raw_data=True,  # For all channels: load RAW data
+                 load_fit_results=True,  # .. load FIT data
+                 load_roi_results=True):  # .. load ROI data
     """
     Data IO for files similar to APS Beamline 13 data format.
     This might be changed later.
 
     Parameters
     ----------
+
     working_directory : str
         path folder
     file_name : str
@@ -643,9 +827,15 @@ def read_hdf_APS(working_directory,
     load_summed_data : bool, optional
         load summed spectrum or not
     load_each_channel : bool, optional
-        load data from each channel or not
-    other_list : list, optional
-        data dumped from suitcase
+        indicates whether to load raw experimental data for each detector channel or not
+    load_raw_data : bool
+        load raw experimental data
+    load_processed_each_channel : bool
+        indicates whether or not to load processed results (fit, roi) for each detector channel
+    load_fit_results :bool
+        load fitting results
+    load_roi_results : bool
+        load results of roi computation
 
     Returns
     -------
@@ -657,6 +847,9 @@ def read_hdf_APS(working_directory,
     data_sets = OrderedDict()
     img_dict = OrderedDict()
 
+    # Empty container for metadata
+    mdata = ScanMetadataXRF()
+
     file_path = os.path.join(working_directory, file_name)
 
     # defined in other_list in config file
@@ -666,12 +859,23 @@ def read_hdf_APS(working_directory,
         dict_sc = {}
 
     with h5py.File(file_path, 'r+') as f:
+
+        # Retrieve metadata if it exists
+        if "xrfmap/scan_metadata" in f:  # Metadata is always loaded
+            metadata = f["xrfmap/scan_metadata"]
+            for key, value in metadata.attrs.items():
+                # Convert ndarrays to lists (they were lists before they were saved)
+                if isinstance(value, np.ndarray):
+                    value = list(value)
+                mdata[key] = value
+
         data = f['xrfmap']
         fname = file_name.split('.')[0]
-        if load_summed_data is True:
+        if load_summed_data and load_raw_data:
             try:
                 # data from channel summed
-                exp_data = np.array(data['detsum/counts'][:, :, 0:spectrum_cut])
+                exp_data = np.array(data['detsum/counts'][:, :, 0:spectrum_cut],
+                                    dtype=np.float32)
                 logger.warning(f"We use spectrum range from 0 to {spectrum_cut}")
                 logger.info(f"Exp. data from h5 has shape of: {exp_data.shape}")
 
@@ -684,7 +888,7 @@ def read_hdf_APS(working_directory,
             except KeyError:
                 print('No data is loaded for detector sum.')
 
-        if 'scalers' in data:
+        if 'scalers' in data:  # Scalers are always loaded if data is available
             det_name = data['scalers/name']
             temp = {}
             for i, n in enumerate(det_name):
@@ -696,7 +900,7 @@ def read_hdf_APS(working_directory,
             if len(dict_sc) != 0:
                 img_dict[f"{fname}_scaler"].update(dict_sc)
 
-        if 'positions' in data:
+        if 'positions' in data:  # Positions are always loaded if data is available
             pos_name = data['positions/name']
             temp = {}
             for i, n in enumerate(pos_name):
@@ -705,6 +909,7 @@ def read_hdf_APS(working_directory,
                 temp[n] = data['positions/pos'].value[i, :]
             img_dict['positions'] = temp
 
+        # TODO: rewrite the algorithm for finding the detector channels (not robust)
         # find total channel:
         channel_num = 0
         for v in list(data.keys()):
@@ -713,12 +918,13 @@ def read_hdf_APS(working_directory,
         channel_num = channel_num-1  # do not consider det_sum
 
         # data from each channel
-        if load_each_channel:
+        if load_each_channel and load_raw_data:
             for i in range(1, channel_num+1):
                 det_name = f"det{i}"
                 file_channel = f"{fname}_det{i}"
                 try:
-                    exp_data_new = np.array(data[f"{det_name}/counts"][:, :, 0:spectrum_cut])
+                    exp_data_new = np.array(data[f"{det_name}/counts"][:, :, 0:spectrum_cut],
+                                            dtype=np.float32)
                     DS = DataSelection(filename=file_channel,
                                        raw_data=exp_data_new)
                     data_sets[file_channel] = DS
@@ -726,60 +932,63 @@ def read_hdf_APS(working_directory,
                 except KeyError:
                     print(f"No data is loaded for {det_name}.")
 
-        for i in range(1, channel_num + 1):
-            det_name = f"det{i}"
-            file_channel = f"{fname}_det{i}"
-            if 'xrf_fit' in data[det_name]:
-                try:
-                    fit_result = get_fit_data(data[det_name]['xrf_fit_name'].value,
-                                              data[det_name]['xrf_fit'].value)
-                    img_dict.update({f"{file_channel}_fit": fit_result})
-                    # also include scaler data
-                    if 'scalers' in data:
-                        img_dict[f"{file_channel}_fit"].update(img_dict[f"{fname}_scaler"])
-                except IndexError:
-                    logger.info(f"No fitting data is loaded for channel {i}.")
-
-            if 'xrf_roi' in data[det_name]:
-                try:
-                    fit_result = get_fit_data(data[det_name]['xrf_roi_name'].value,
-                                              data[det_name]['xrf_roi'].value)
-                    img_dict.update({f"{file_channel}_roi": fit_result})
-                    # also include scaler data
-                    if 'scalers' in data:
-                        img_dict[f"{file_channel}_roi"].update(img_dict[f"{fname}_scaler"])
-                except IndexError:
-                    logger.info(f"No ROI data is loaded for channel {i}.")
-
-        if 'roimap' in data:
-            if 'sum_name' in data['roimap']:
-                det_name = data['roimap/sum_name']
-                temp = {}
-                for i, n in enumerate(det_name):
-                    temp[n] = data['roimap/sum_raw'].value[:, :, i]
-                    # bad points on first one
+        if load_processed_each_channel:
+            for i in range(1, channel_num + 1):
+                det_name = f"det{i}"
+                file_channel = f"{fname}_det{i}"
+                if 'xrf_fit' in data[det_name] and load_fit_results:
                     try:
-                        temp[n][0, 0] = temp[n][1, 0]
+                        fit_result = get_fit_data(data[det_name]['xrf_fit_name'].value,
+                                                  data[det_name]['xrf_fit'].value)
+                        img_dict.update({f"{file_channel}_fit": fit_result})
+                        # also include scaler data
+                        if 'scalers' in data:
+                            img_dict[f"{file_channel}_fit"].update(img_dict[f"{fname}_scaler"])
                     except IndexError:
-                        temp[n][0, 0] = temp[n][0, 1]
-                img_dict[f"{fname}_roi"] = temp
-                # also include scaler data
-                if 'scalers' in data:
-                    img_dict[f"{fname}_roi"].update(img_dict[f"{fname}_scaler"])
+                        logger.info(f"No fitting data is loaded for channel {i}.")
 
-            if 'det_name' in data['roimap']:
-                det_name = data['roimap/det_name']
-                temp = {}
-                for i, n in enumerate(det_name):
-                    temp[n] = data['roimap/det_raw'].value[:, :, i]
+                if 'xrf_roi' in data[det_name] and load_roi_results:
                     try:
-                        temp[n][0, 0] = temp[n][1, 0]
+                        fit_result = get_fit_data(data[det_name]['xrf_roi_name'].value,
+                                                  data[det_name]['xrf_roi'].value)
+                        img_dict.update({f"{file_channel}_roi": fit_result})
+                        # also include scaler data
+                        if 'scalers' in data:
+                            img_dict[f"{file_channel}_roi"].update(img_dict[f"{fname}_scaler"])
                     except IndexError:
-                        temp[n][0, 0] = temp[n][0, 1]
-                img_dict[f"{fname}_roi_each"] = temp
+                        logger.info(f"No ROI data is loaded for channel {i}.")
+
+        # Prepare the following code for possible deletion in the future
+        #               (we don't use ROI maps recorded during the experiment)
+        # if 'roimap' in data:
+        #    if 'sum_name' in data['roimap']:
+        #        det_name = data['roimap/sum_name']
+        #        temp = {}
+        #        for i, n in enumerate(det_name):
+        #            temp[n] = data['roimap/sum_raw'].value[:, :, i]
+        #            # bad points on first one
+        #            try:
+        #                temp[n][0, 0] = temp[n][1, 0]
+        #            except IndexError:
+        #                temp[n][0, 0] = temp[n][0, 1]
+        #        img_dict[f"{fname}_roi"] = temp
+        #        # also include scaler data
+        #        if 'scalers' in data:
+        #            img_dict[f"{fname}_roi"].update(img_dict[f"{fname}_scaler"])
+        #
+        #    if 'det_name' in data['roimap']:
+        #        det_name = data['roimap/det_name']
+        #        temp = {}
+        #        for i, n in enumerate(det_name):
+        #            temp[n] = data['roimap/det_raw'].value[:, :, i]
+        #            try:
+        #                temp[n][0, 0] = temp[n][1, 0]
+        #            except IndexError:
+        #                temp[n][0, 0] = temp[n][0, 1]
+        #        img_dict[f"{fname}_roi_each"] = temp
 
         # read fitting results from summed data
-        if 'xrf_fit' in data['detsum']:
+        if 'xrf_fit' in data['detsum'] and load_summed_data and load_fit_results:
             try:
                 fit_result = get_fit_data(data['detsum']['xrf_fit_name'].value,
                                           data['detsum']['xrf_fit'].value)
@@ -789,7 +998,7 @@ def read_hdf_APS(working_directory,
             except (IndexError, KeyError):
                 logger.info('No fitting data is loaded for channel summed data.')
 
-        if 'xrf_roi' in data['detsum']:
+        if 'xrf_roi' in data['detsum'] and load_summed_data and load_roi_results:
             try:
                 fit_result = get_fit_data(data['detsum']['xrf_roi_name'].value,
                                           data['detsum']['xrf_roi'].value)
@@ -799,10 +1008,10 @@ def read_hdf_APS(working_directory,
             except (IndexError, KeyError):
                 logger.info('No ROI data is loaded for summed data.')
 
-    return img_dict, data_sets
+    return img_dict, data_sets, mdata
 
 
-def render_data_to_gui(runid, *, file_overwrite_existing=False):
+def render_data_to_gui(runid, *, create_each_det=False, working_directory=None, file_overwrite_existing=False):
     """
     Read data from databroker and save to Atom class which GUI can take.
 
@@ -812,6 +1021,11 @@ def render_data_to_gui(runid, *, file_overwrite_existing=False):
     ----------
     runid : int
         id number for given run
+    create_each_det : bool
+        True: load data from all detector channels
+        False: load only the sum of all channels
+    working_directory : str
+        path to the directory where data files are saved
     file_overwrite_existing : bool
         True: overwrite data file if it exists
         False: create unique file name by adding version number
@@ -825,11 +1039,17 @@ def render_data_to_gui(runid, *, file_overwrite_existing=False):
     # Don't create unique file name if the existing file is to be overwritten
     fname_add_version = not file_overwrite_existing
 
+    # Create file name here, so that working directory may be attached to the file name
+    prefix = 'scan2D_'
+    fname = f"{prefix}{runid}.h5"
+    if working_directory:
+        fname = os.path.join(working_directory, fname)
+
     data_from_db = fetch_data_from_db(runid,
+                                      fpath=fname,
                                       fname_add_version=fname_add_version,
                                       file_overwrite_existing=file_overwrite_existing,
-                                      # Always load data from all detectors
-                                      create_each_det=True,
+                                      create_each_det=create_each_det,
                                       # Always create data file (processing results
                                       #   are going to be saved in the file)
                                       output_to_file=True)
@@ -850,6 +1070,7 @@ def render_data_to_gui(runid, *, file_overwrite_existing=False):
     data_out = data_from_db[0]['dataset']
     fname = data_from_db[0]['file_name']
     detector_name = data_from_db[0]['detector_name']
+    scan_metadata = data_from_db[0]['metadata']
 
     # Create file name for the 'sum' dataset ('file names' are used as dictionary
     #   keys in data storage containers, as channel labels in plot legends,
@@ -902,7 +1123,7 @@ def render_data_to_gui(runid, *, file_overwrite_existing=False):
         scaler_tmp[v] = data_out['scaler_data'][:, :, i]
     img_dict[fname_no_ext+'_scaler'] = scaler_tmp
     logger.info("Data loading: scaler data are loaded successfully.")
-    return img_dict, data_sets, fname, detector_name
+    return img_dict, data_sets, fname, detector_name, scan_metadata
 
 
 def retrieve_data_from_hdf_suitcase(fpath):
@@ -957,6 +1178,9 @@ def read_MAPS(working_directory,
     data_sets = OrderedDict()
     img_dict = OrderedDict()
 
+    # Empty container for metadata
+    mdata = ScanMetadataXRF()
+
     #  cut off bad point on the last position of the spectrum
     # bad_point_cut = 0
 
@@ -1045,7 +1269,7 @@ def read_MAPS(working_directory,
     #                               data[detID]['xrf_fit'].value)
     #     img_dict.update({fname+'_fit': fit_result})
 
-    return img_dict, data_sets
+    return img_dict, data_sets, mdata
 
 
 def get_roi_sum(namelist, data_range, data):
@@ -1547,27 +1771,6 @@ def create_movie(data, fname='demo.mp4', dpi=100, cmap='jet',
     ani.save(fname, writer=writer, dpi=dpi)
 
 
-def print_image(fig):
-    """
-    Print function used at beamline only.
-    """
-    if db is not None:
-        hdr = db[-1]
-        if hdr.start.beamline_id == 'HXN':
-            current_dir = os.path.dirname(os.path.realpath(__file__))
-            config_file = 'hxn_pv_config.json'
-            config_path = sep_v.join(current_dir.split(sep_v)[:-2]+['configs', config_file])
-            with open(config_path, 'r') as json_data:
-                config_data = json.load(json_data)
-            fpath = config_data.get('print_path', None)
-            if fpath is None:
-                fpath = '/home/xf03id/Desktop/temp.png'
-            fig.savefig(fpath,  bbox_inches='tight', pad_inches=4)
-            os.system(config_data['print_command'])
-        else:
-            print('Printer is not set up yet.')
-
-
 def spec_to_hdf(wd, spec_file, spectrum_file, output_file, img_shape,
                 ic_name=None, x_name=None, y_name=None):
     """
