@@ -9,14 +9,14 @@ import re
 import math
 from collections import OrderedDict, deque
 import multiprocessing
+import multiprocessing.pool
 import h5py
 import matplotlib.pyplot as plt
 import matplotlib.animation as animation
 import json
-# from scipy.optimize import nnls
-from scipy.interpolate import interp1d, interp2d
 import lmfit
-# from enaml.qt.qt_application import QApplication
+import platform
+from distutils.version import LooseVersion
 
 from atom.api import Atom, Str, observe, Typed, Int, List, Dict, Float, Bool
 from skbeam.core.fitting.xrf_model import (ModelSpectrum, update_parameter_dict,
@@ -33,10 +33,12 @@ from .guessparam import (calculate_profile, fit_strategy_list,
                          trim_escape_peak, define_range, get_energy,
                          get_Z, PreFitStatus, ElementController,
                          update_param_from_element)
-from .fileio import save_fitdata_to_hdf, output_data, output_data_to_tiff
+from .fileio import save_fitdata_to_hdf, output_data
 
-from .utils import (gaussian_sigma_to_fwhm, gaussian_fwhm_to_sigma,
-                    gaussian_max_to_area, gaussian_area_to_max)
+from ..core.utils import (gaussian_sigma_to_fwhm, gaussian_fwhm_to_sigma,
+                          gaussian_max_to_area, gaussian_area_to_max)
+
+from ..core.quant_analysis import ParamQuantEstimation
 
 import logging
 logger = logging.getLogger()
@@ -57,6 +59,8 @@ class Fit1D(Atom):
         dict of 2D array map for each fitted element
     map_interpolation : bool
         option to interpolate the 2D map according to x,y position or not
+        Interpolation is performed only before exporting data
+        (currently .tiff or .txt)
     hdf_path : str
         path to hdf file
     hdf_name : str
@@ -66,6 +70,9 @@ class Fit1D(Atom):
     default_parameters = Dict()
     param_dict = Dict()
 
+    img_dict = Dict()
+    data_sets = Typed(OrderedDict)
+
     element_list = List()
     data_sets = Dict()
     data_all = Typed(np.ndarray)
@@ -97,10 +104,24 @@ class Fit1D(Atom):
     cal_spectrum = Dict()
 
     #  attributes used by the ElementEdit window
-    # selected_element = Str()
+    selected_element = Str()
     selected_index = Int()
     elementinfo_list = List()
 
+    # The variable contains the image title displayed in "Element Map' tab.
+    #   The variable is synchronized with the identical variable in 'DrawImageAdvanced' class.
+    img_title = Str()
+    # Reference to the dictionary that contains dataset currently displayed in 'Element Map' tab
+    #   The variable is synchronized with the identical variable in 'DrawImageAdvanced' class
+    dict_to_plot = Dict()
+
+    # The variable is replicating the identical variable in 'DrawImageAdvanced' class
+    #   True - quantitative normalization is ON, False - OFF
+    quantitative_normalization = Bool()
+    # The reference to the object holding parameters for quantitative normalization
+    #   The variable is synchronized to the identical variable in 'DrawImageAdvanced' class
+    param_quant_analysis = Typed(object)
+
     function_num = Int(0)
     nvar = Int(0)
     chi2 = Float(0.0)
@@ -137,8 +158,6 @@ class Fit1D(Atom):
     pixel_fit_method = Int(0)
     param_q = Typed(object)
 
-    x_data = Typed(np.ndarray)
-    y_data = Typed(np.ndarray)
     result_map = Dict()
     map_interpolation = Bool(False)
     hdf_path = Str()
@@ -151,6 +170,9 @@ class Fit1D(Atom):
     # Reference to GuessParamModel object
     param_model = Typed(object)
 
+    # Reference to FileIOMOdel
+    io_model = Typed(object)
+
     # Fields for updating user defined peak parameters
     add_userpeak_energy = Float(0.0)
     add_userpeak_fwhm = Float(0.0)
@@ -163,7 +185,29 @@ class Fit1D(Atom):
     name_userpeak_dsigma = Str()
     name_userpeak_area = Str()
 
-    def __init__(self, param_model, *args, **kwargs):
+    # Quantitative analysis: used during estimation step
+    param_quant_estimation = ParamQuantEstimation()
+    # *** The following two references are used exclusively to update the list of standards ***
+    # ***   in the dialog box. ***
+    qe_param_built_in_ref = Typed(object)
+    qe_param_custom_ref = Typed(object)
+    # The following reference used to track the selected standard while the selection
+    #   dialog box is open. Once the dialog box is opened again, the reference becomes
+    #   invalid, since the descriptions of the standards are reloaded from files.
+    qe_standard_selected_ref = Typed(object)
+    # Keep the actual copy of the selected standard. The copy is used to keep information
+    #   on the selected standard while descriptions are reloaded (references become invalid).
+    qe_standard_selected_copy = Typed(object)
+    # *** The following fields are used exclusively to store input values ***
+    # ***   for the 'SaveQuantCalibration' dialog box. ***
+    # *** The fields are not guaranteed to have valid values at any other time. ***
+    qe_standard_path_name = Str()
+    qe_standard_file_name = Str()
+    qe_standard_distance = Str("0.0")
+    qe_standard_overwrite_existing = Bool(False)
+    qe_standard_data_preview = Str()
+
+    def __init__(self, param_model, io_model, *args, **kwargs):
         self.working_directory = kwargs['working_directory']
         self.result_folder = kwargs['working_directory']
         self.default_parameters = kwargs['default_parameters']
@@ -174,6 +218,9 @@ class Fit1D(Atom):
         # Reference to GuessParamModel object
         self.param_model = param_model
 
+        # Reference to FileIOMOdel
+        self.io_model = io_model
+
         self.EC = ElementController()
         self.pileup_data = {'element1': 'Si_K',
                             'element2': 'Si_K',
@@ -190,6 +237,9 @@ class Fit1D(Atom):
         self.roi_sum_opt['low'] = 0.0
         self.roi_sum_opt['high'] = 10.0
 
+        self.qe_standard_selected_ref = None
+        self.qe_standard_selected_copy = None
+
     def result_folder_changed(self, change):
         """
         Observer function to be connected to the fileio model
@@ -240,10 +290,17 @@ class Fit1D(Atom):
             This is the dictionary that gets passed to a function
             with the @observe decorator
         """
-        img_dict = change['value']
-        _key = [k for k in img_dict.keys() if 'scaler' in k]
+        self.img_dict = change['value']
+        _key = [k for k in self.img_dict.keys() if 'scaler' in k]
         if len(_key) != 0:
-            self.scaler_keys = sorted(img_dict[_key[0]].keys())
+            self.scaler_keys = sorted(self.img_dict[_key[0]].keys())
+
+    def data_sets_update(self, change):
+        """
+        Observer function to be connected to the fileio model
+        in the top-level gui.py startup
+        """
+        self.data_sets = change['value']
 
     def scaler_index_update(self, change):
         """
@@ -258,6 +315,26 @@ class Fit1D(Atom):
         """
         self.scaler_index = change['value']
 
+    def img_title_update(self, change):
+        r"""Observer function. Sets ``img_title`` field to the same value as
+        the identical variable in ``DrawImageAdvanced`` class"""
+        self.img_title = change['value']
+
+    def dict_to_plot_update(self, change):
+        r"""Observer function. Sets ``dict_to_plot`` field to the same value as
+        the identical variable in ``DrawImageAdvanced`` class"""
+        self.dict_to_plot = change['value']
+
+    def quantitative_normalization_update(self, change):
+        r"""Observer function. Sets ``quantitative_normalization`` field to the same value as
+        the identical variable in ``DrawImageAdvanced`` class"""
+        self.quantitative_normalization = change['value']
+
+    def param_quant_analysis_update(self, change):
+        r"""Observer function. Sets ``param_quant_analysis`` field to the same value as
+        the identical variable in ``DrawImageAdvanced`` class"""
+        self.param_quant_analysis = change['value']
+
     def energy_bound_high_update(self, change):
         """
         Observer function that connects 'param_model' (GuessParamModel)
@@ -274,6 +351,38 @@ class Fit1D(Atom):
         """
         self.param_dict['non_fitting_values']['energy_bound_low']['value'] = change['value']
 
+    def update_selected_index(self, selected_element=None,
+                              element_list_new=None):
+
+        if selected_element is None:
+            # Currently selected element
+            element = self.selected_element
+        else:
+            # Selected element (probably saved before update of the element list)
+            element = selected_element
+
+        if element_list_new is None:
+            # Current element list
+            element_list = self.element_list
+        else:
+            # Future element list (before it is updated)
+            element_list = element_list_new
+
+        if not element_list:
+            # Empty element list
+            ind = 0
+        else:
+            try:
+                # Combo-box has additional element 'Select element'
+                ind = element_list.index(element) + 1
+            except ValueError:
+                # Element is not found (was deleted), so deselect the element
+                ind = 0
+        if ind == self.selected_index:
+            # We want the content to update (deselect, then select again)
+            self.selected_index = 0
+        self.selected_index = ind
+
     @observe('selected_index')
     def _selected_element_changed(self, change):
         if change['value'] > 0:
@@ -281,18 +390,43 @@ class Fit1D(Atom):
             if ind_sel >= len(self.element_list):
                 ind_sel = len(self.element_list) - 1
                 self.selected_index = ind_sel + 1  # Change the selection as well
-            selected_element = self.element_list[ind_sel]
-            if len(selected_element) <= 4:
-                element = selected_element.split('_')[0]
+            self.selected_element = self.element_list[ind_sel]
+            if len(self.selected_element) <= 4:
+                element = self.selected_element.split('_')[0]
                 self.elementinfo_list = sorted([e for e in list(self.param_dict.keys())
                                                 if (element+'_' in e) and  # error between S_k or Si_k
                                                 ('pileup' not in e)])  # Si_ka1 not Si_K
                 logger.info(f"Element line info: {self.elementinfo_list}")
             else:
-                element = selected_element  # for pileup peaks
+                element = self.selected_element  # for pileup peaks
                 self.elementinfo_list = sorted([e for e in list(self.param_dict.keys())
                                                 if element.replace('-', '_') in e])
                 logger.info(f"User defined or pileup peak info: {self.elementinfo_list}")
+        else:
+            self.elementinfo_list = []
+
+    @observe('qe_standard_distance')
+    def _qe_standard_distance_changed(self, change):
+        try:
+            d = float(change["value"])
+        except Exception:
+            d = None
+        if d <= 0.0:
+            d = None
+        self.param_quant_estimation.set_distance_to_sample_in_data_dict(distance_to_sample=d)
+        # Change preview if distance value changed
+        self.qe_standard_data_preview = \
+            self.param_quant_estimation.get_fluorescence_data_dict_text_preview()
+
+    def get_qe_standard_distance_as_float(self):
+        r"""Return distance from sample as positive float or None"""
+        try:
+            d = float(self.qe_standard_distance)
+        except Exception:
+            d = None
+        if (d is not None) and (d <= 0.0):
+            d = None
+        return d
 
     def _compute_fwhm_base(self):
         # Computes 'sigma' value based on default parameters and peak energy (for Userpeaks)
@@ -302,38 +436,7 @@ class Fit1D(Atom):
 
         sigma = gaussian_fwhm_to_sigma(self.param_model.default_parameters["fwhm_offset"]["value"])
 
-        # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
-        # The error was detected in scikit-beam, but not corrected
-        # Note: Corrections are required in the function ``element_peak_xrf``
-        #   from file ``xrf_model.py`` (package Scikit-Beam)
-        # The error: the original center position for the peak was used to
-        #   compute sigma of the Gaussian approximation of the peak. It almost not matter
-        #   for element peaks, since they don't move along the energy axis. But it is
-        #   invalid for Userpeaks, which are placed at a fixed position (5 keV) and then
-        #   moved to the desired position by changing ``delta_center``. Sigma depends
-        #   on actual center position, so ``delta_center`` should be taken into account.
-        #
-        # The original code for computing Gaussian function:
-        # def get_sigma(center):
-        #     temp_val = 2 * np.sqrt(2 * np.log(2))
-        #     return np.sqrt((fwhm_offset / temp_val) ** 2 + center * epsilon * fwhm_fanoprime)
-        #
-        # x = e_offset + x * e_linear + x ** 2 * e_quadratic
-        #
-        # return gaussian(x, area, center + delta_center,
-        #                 delta_sigma + get_sigma(center)) * ratio * ratio_adjust
-        #
-        # To correct the error, the last statement should be changed to the following:
-        # return gaussian(x, area, center + delta_center,
-        #                 delta_sigma + get_sigma(center + delta_center)) * ratio * ratio_adjust
-
-        # Temporary fix for the error mentioned above (to keep peak hight consistent as it moves
-        #   along the enery axis, computation of sigma must be identical here and in
-        #   the function ``element_peak_xrf``
-        sigma_sqr = 5.0  # center (ignore ``delta_sigma``)
-        # The next line is correct (but may be enabled only if the bug in scikit-beam is corrected)
-        # sigma_sqr = self.param_dict[self.name_userpeak_dcenter]["value"] + 5.0  # center
-
+        sigma_sqr = self.param_dict[self.name_userpeak_dcenter]["value"] + 5.0  # center
         sigma_sqr *= self.param_model.default_parameters["non_fitting_values"]["epsilon"]  # epsilon
         sigma_sqr *= self.param_model.default_parameters["fwhm_fanoprime"]["value"]  # fanoprime
         sigma_sqr += sigma * sigma  # We have computed the expression under sqrt
@@ -487,6 +590,28 @@ class Fit1D(Atom):
                      f"          Energy: {self.add_userpeak_energy} keV\n"
                      f"          FWHM: {self.add_userpeak_fwhm} keV")
 
+    def update_userpeak_controls(self):
+        """
+        The function should be called right after adding a userpeak to update the fields
+        for userpeak energy and fwhm. Uses data for the currently selected Userpeak
+        (the peak should be selected at the time of creation!!!)
+        """
+        if (self.name_userpeak_dcenter not in self.param_dict) or \
+                (self.name_userpeak_dsigma not in self.param_dict):
+            return
+        # Set energy
+        v_center = self.param_dict[self.name_userpeak_dcenter]["value"]
+        v_energy = v_center + 5.0
+        v_energy = round(v_energy, 3)
+        self.add_userpeak_energy = v_energy
+        # Set fwhm
+        fwhm_base = self._compute_fwhm_base()
+        v_dsigma = self.param_dict[self.name_userpeak_dsigma]["value"]
+        v_dfwhm = gaussian_sigma_to_fwhm(v_dsigma)
+        v_fwhm = v_dfwhm + fwhm_base
+        v_fwhm = round(v_fwhm, 5)
+        self.add_userpeak_fwhm = v_fwhm
+
     def keep_size(self):
         """Keep the size of deque as 2.
         """
@@ -504,6 +629,7 @@ class Fit1D(Atom):
         """
         with open(param_path, 'r') as json_data:
             self.default_parameters = json.load(json_data)
+
         #  use queue to save the status of parameters
         self.param_q.append(copy.deepcopy(self.default_parameters))
         self.keep_size()
@@ -524,16 +650,26 @@ class Fit1D(Atom):
         """
         Update param_dict with default parameters, also update element list.
         """
-        self.param_dict = copy.deepcopy(self.default_parameters)
+        # Save currently selected element name
+        selected_element = self.selected_element
+        self.selected_index = 0
 
-        element_list = self.param_dict['non_fitting_values']['element_list']
-        self.element_list = [e.strip(' ') for e in element_list.split(',')]
+        element_list = self.default_parameters['non_fitting_values']['element_list']
+        element_list = [e.strip(' ') for e in element_list.split(',')]
+        element_list = [_ for _ in element_list if _]  # Get rid of empty strings in the list
+        self.element_list = element_list
+
+        self.param_dict = copy.deepcopy(self.default_parameters)
 
         # show the list of elements on add/remove window
         self.EC.delete_all()
         self.create_EC_list(self.element_list)
         self.update_name_list()
 
+        # Update the index in case the selected emission line disappeared from the list
+        self.update_selected_index(selected_element=selected_element,
+                                   element_list_new=element_list)
+
         # global parameters
         # for GUI purpose only
         # if we do not clear the list first, there is not update on the GUI
@@ -720,7 +856,8 @@ class Fit1D(Atom):
             y0 = self.y0 - self.bg
 
         t0 = time.time()
-        self.fit_info = 'Summed spectrum fitting is in process.'
+        self.fit_info = "Spectrum fitting of the sum spectrum (incident energy "\
+                        f"{self.param_dict['coherent_sct_energy']['value']})."
         # app.processEvents()
         # logger.info('-------- '+self.fit_info+' --------')
 
@@ -739,6 +876,21 @@ class Fit1D(Atom):
                 self.fit_data(self.x0, y0)
                 self.update_param_with_result()
 
+                # The following is a patch for rare cases when fitting results in negative
+                #   areas for some emission lines. These are typically non-existent lines, but
+                #   they should not be automatically eliminated from the list. To prevent
+                #   elimination, set the area to some small positive value.
+                for key, val in self.param_dict.items():
+                    if key.endswith("_area") and val["value"] <= 0.0:
+                        _small_value_for_area = 0.1
+                        logger.warning(
+                            f"Fitting resulted in negative value for '{key}' ({val['value']}). \n"
+                            f"    In order to continue using the emission line in future computations, "
+                            f"the fitted area is set to a small value ({_small_value_for_area}).\n    Delete "
+                            f"the emission line from the list if you know it is not present in "
+                            f"the sample.")
+                        val["value"] = _small_value_for_area  # Some small number
+
                 #  calculate r2
                 self.r2 = cal_r2(y0, self.fit_y)
                 self.assign_fitting_result()
@@ -803,9 +955,11 @@ class Fit1D(Atom):
         elif self.pixel_fit_method == 1:
             pixel_fit = 'nonlinear'
 
-        logger.info('-------- Fitting of single pixels starts. --------')
+        logger.info("-------- Fitting of single pixels starts (incident_energy "
+                    f"{self.param_dict['coherent_sct_energy']['value']} keV) --------")
         t0 = time.time()
         self.pixel_fit_info = 'Pixel fitting is in process.'
+
         # app.processEvents()
         self.result_map, calculation_info = single_pixel_fitting_controller(
             self.data_all,
@@ -821,18 +975,6 @@ class Fit1D(Atom):
         t1 = time.time()
         logger.info('Time used for pixel fitting is : {}'.format(t1-t0))
 
-        try:
-            with h5py.File(self.hdf_path, 'r') as f:
-                self.x_data = np.array(f['xrfmap/positions/pos'][0, :, :])
-                self.y_data = np.array(f['xrfmap/positions/pos'][1, :, :])
-
-            # this is consistent with fileIO that x data is fliped.
-            self.x_data = np.fliplr(self.x_data)
-        except KeyError:
-            pass
-        except ValueError:
-            pass
-
         #  get fitted spectrum and save them to figs
         if self.save_point is True:
             self.pixel_fit_info = 'Saving output ...'
@@ -871,9 +1013,11 @@ class Fit1D(Atom):
             self.save2Dmap_to_hdf(calculation_info=calculation_info, pixel_fit=pixel_fit)
             self.pixel_fit_info = 'Pixel fitting is done!'
             # app.processEvents()
-            logger.info('-------- Fitting of single pixels is done! --------')
         except ValueError:
             logger.warning('Fitting result can not be saved to h5 file.')
+        except IOError as ex:
+            logger.warning(f"{ex}")
+        logger.info('-------- Fitting of single pixels is done! --------')
 
     def save_pixel_fitting_to_db(self):
         """Save fitting results to analysis store
@@ -895,19 +1039,6 @@ class Fit1D(Atom):
         pixel_fit : str
             If nonlinear is chosen, more information needs to be saved.
         """
-        if self.map_interpolation is True:
-            logger.info('Interpolating image... ')
-            rangex = self.x_data[0, :]
-            rangey = self.y_data[:, 0]
-            start_x = rangex[0]
-            start_y = rangey[0]
-            dimv = self.data_all.shape
-            for k, v in six.iteritems(self.result_map):
-                shapev = [dimv[1], dimv[0]]  # horizontal first, then vertical, different from dim in numpy
-                interp_d = interp1d_scan(shapev, rangex, rangey, start_x, start_y,
-                                         self.x_data, self.y_data, v)
-                interp_d[np.isnan(interp_d)] = 0
-                self.result_map[k] = interp_d
 
         prefix_fname = self.hdf_name.split('.')[0]
         if len(prefix_fname) == 0:
@@ -930,6 +1061,9 @@ class Fit1D(Atom):
         # Update GUI so that results can be seen immediately
         self.fit_img[fit_name] = self.result_map
 
+        if not os.path.isfile(self.hdf_path):
+            raise IOError(f"File '{self.hdf_path}' does not exist. Data is not saved to HDF5 file.")
+
         save_fitdata_to_hdf(self.hdf_path, self.result_map, datapath=inner_path)
 
         # output error
@@ -950,6 +1084,24 @@ class Fit1D(Atom):
             # save to hdf again, this is not optimal
             self.save2Dmap_to_hdf()
 
+    def get_latest_single_pixel_fitting_data(self):
+        r"""
+        Returns the latest results of single pixel fitting. The returned results include
+        the name of the scaler (None if no scaler is selected) and the dictionary of
+        the computed XRF maps for selected emission lines.
+        """
+
+        # Find the selected scaler name. Scaler is None if not scaler is selected.
+        scaler_name = None
+        if self.scaler_index > 0:
+            scaler_name = self.scaler_keys[self.scaler_index-1]
+
+        # Result map. If 'None', then no fitting results exists.
+        #   Single pixel fitting must be run
+        result_map = self.result_map.copy()
+
+        return result_map, scaler_name
+
     def output_2Dimage(self, to_tiff=True):
         """Read data from h5 file and save them into either tiff or txt.
 
@@ -960,36 +1112,44 @@ class Fit1D(Atom):
         """
         scaler_v = None
         _post_name_folder = "_".join(self.data_title.split('_')[:-1])
-        _name_each_file = "_".join(self.data_title.split('_')[1:])
         if self.scaler_index > 0:
             scaler_v = self.scaler_keys[self.scaler_index-1]
-            logger.info('*** Data will be saved with NORMALIZATION '
-                        'from {} ***'.format(scaler_v))
+            logger.info(f"*** NORMALIZED data is saved. Scaler: '{scaler_v}' ***")
+
         if to_tiff:
-            output_n = 'output_tiff_' + _post_name_folder
-            output_full_name = os.path.join(self.result_folder, output_n)
-            # still keep the function of reading data from hdf and saving,
-            # for cases that user wants to output current data in hdf
-            # without rerun fitting again
-            if os.path.exists(self.hdf_path):
-                output_data(self.hdf_path, output_full_name,
-                            norm_name=scaler_v)
-            else:
-                output_data_to_tiff(self.result_map, output_full_name,
-                                    name_append=_name_each_file,
-                                    norm_name=scaler_v)
-            logger.info('Done with saving data {} to tiff files.'.format(output_n))
+            dir_prefix = "output_tiff_"
+            file_format = "tiff"
         else:
-            output_n = 'output_txt_' + _post_name_folder
-            output_full_name = os.path.join(self.result_folder, output_n)
-            if os.path.exists(self.hdf_path):
-                output_data(self.hdf_path, output_full_name,
-                            file_format='txt', norm_name=scaler_v)
-            else:
-                output_data_to_tiff(self.result_map, output_full_name,
-                                    name_append=_name_each_file,
-                                    file_format='txt', norm_name=scaler_v)
-            logger.info('Done with saving data {} to txt files.'.format(output_n))
+            dir_prefix = "output_txt_"
+            file_format = "txt"
+
+        output_n = dir_prefix + _post_name_folder
+        output_dir = os.path.join(self.result_folder, output_n)
+
+        # self.img_dict contains ALL loaded datasets, including a separate "positions" dataset
+        if "positions" in self.img_dict:
+            positions_dict = self.img_dict["positions"]
+        else:
+            positions_dict = {}
+
+        # Scalers are located in a separate dataset in 'img_dict'. They are also referenced
+        #   in each '_fit' dataset (and in the selected dataset 'self.dict_to_plot')
+        #   The list of scaler names is used to avoid attaching the detector channel name
+        #   to file names that contain scaler data (scalers typically do not depend on
+        #   the selection of detector channels.
+        scaler_dsets = [_ for _ in self.img_dict.keys() if re.search(r"_scaler$", _)]
+        if scaler_dsets:
+            scaler_name_list = list(self.img_dict[scaler_dsets[0]].keys())
+        else:
+            scaler_name_list = None
+
+        output_data(output_dir=output_dir,
+                    interpolate_to_uniform_grid=self.map_interpolation,
+                    dataset_name=self.img_title, quant_norm=self.quantitative_normalization,
+                    param_quant_analysis=self.param_quant_analysis,
+                    dataset_dict=self.dict_to_plot, positions_dict=positions_dict,
+                    file_format=file_format, scaler_name=scaler_v,
+                    scaler_name_list=scaler_name_list)
 
     def save_result(self, fname=None):
         """
@@ -1623,7 +1783,7 @@ def fit_per_line_nnls(row_num, data,
         fitting values for all the elements at a given row. Background is
         calculated as a summed value. Also residual is included.
     """
-    logger.info('Row number at {}'.format(row_num))
+    logger.debug(f"Row number at {row_num}")
     out = []
     bg_sum = 0
     for i in range(data.shape[0]):
@@ -1677,7 +1837,32 @@ def fit_pixel_multiprocess_nnls(exp_data, matv, param,
     num_processors_to_use = multiprocessing.cpu_count()
 
     logger.info('cpu count: {}'.format(num_processors_to_use))
-    pool = multiprocessing.Pool(num_processors_to_use)
+
+    # TODO: find better permanent solution for issue with multiprocessing on Catalina MacOS
+    # =======================================================================================
+    # The following is a temporary patch for the issue with multiprocessing on Catalina MacOS.
+    # The issue is solved by disabling multiprocessing if current OS is Catalina MacOS.
+    # The code will change when the better solution is found.
+
+    # Check if the OS is MacOS Catalina (10.15) or older
+    disable_multiprocessing = False
+    try:
+        os_version = platform.mac_ver()[0]
+        # 'os_version' is an empty string if the system is not MacOS
+        # os_version has format similar to '10.15.3', MacOS Catalina is 10.15
+        if os_version and (LooseVersion(os_version) >= LooseVersion("10.15")):
+            disable_multiprocessing = True
+    except Exception as ex:
+        logger.error(f"Error occurred while checking the version of MacOS: {ex}")
+
+    if disable_multiprocessing:
+        pool = multiprocessing.pool.ThreadPool(num_processors_to_use)
+        logger.warning("Multiprocessing is currently not supported when running PyXRF in MacOS Catalina. "
+                       "Computations are executed in multithreading mode instead.")
+    else:
+        pool = multiprocessing.Pool(num_processors_to_use)
+        logger.info("Computations are executed in multiprocessing mode.")
+
     n_data, n_feature = matv.shape
 
     if lambda_reg > 0:
@@ -1879,7 +2064,7 @@ def single_pixel_fitting_controller(input_data, parameter,
     """
     param = copy.deepcopy(parameter)
     if incident_energy is not None:
-        param['coherent_sct_amplitude']['value'] = incident_energy
+        param['coherent_sct_energy']['value'] = incident_energy
     # cut data into proper range
     x, exp_data, fit_range = get_cutted_spectrum_in3D(input_data,
                                                       param['non_fitting_values']['energy_bound_low']['value'],
@@ -1892,6 +2077,18 @@ def single_pixel_fitting_controller(input_data, parameter,
     elist = [e.strip(' ') for e in elist]
     e_select, matv, e_area = construct_linear_model(x, param, elist)
 
+    # The initial list of elines may contain lines that are not activated for the incident beam
+    #   energy. This always happens for at least one line when batches of XRF scans obtained for
+    #   the range of beam energies are fitted for the same selection of emission lines to generate
+    #   XANES amps. In such experiments, the line of interest is typically not activated at the
+    #   lower energies of the band. It is impossible to include non-activated lines in the linear
+    #   model. In order to make processing results consistent throughout the batch (contain the
+    #   same set of emission lines), the non-activated lines are represented by maps filled with zeros.
+    elist_non_activated = list(set(elist) - set(e_select))
+    if elist_non_activated:
+        logger.warning("Some of the emission lines in the list are not activated: "
+                       f"{elist_non_activated} at {param['coherent_sct_energy']['value']} keV.")
+
     if comp_elastic_combine is True:
         e_select = e_select[:-1]
         e_select[-1] = 'comp_elastic'
@@ -1947,6 +2144,15 @@ def single_pixel_fitting_controller(input_data, parameter,
                                                                           fit_results,
                                                                           matv/matrix_norm)
 
+    # Generate 'zero' maps for the emission lines that were not activated
+    for eline in elist_non_activated:
+        result_map[eline] = np.zeros(shape=exp_data.shape[0:2])
+
+    # Compute total count for each pixel
+    total_count = np.sum(input_data, axis=2)
+    # Save the map as 'total_cnt'
+    result_map['total_cnt'] = total_count
+
     calculation_info = dict()
     if error_map is not None:
         calculation_info['error_map'] = error_map
@@ -2149,87 +2355,3 @@ def get_cs(elemental_line, eng=12, norm=False, round_n=2):
         return np.around(sumv/e.csb(eng)[name_label], round_n)
     else:
         return np.around(sumv, round_n)
-
-
-def fly2d_grid(dimv, rangex, rangey, start_x, start_y,
-               x_data=None, y_data=None):
-    """Get ideal gridded points for a 2D flyscan.
-    .. warning: dimv[dimh, dimv] here is different from tradition dim in python which is [dimv, dimh].
-
-    Parameters
-    ----------
-    dimv : array
-        horizontal first, then vertical.
-    rangex : array, list
-        range along horizontal direction.
-    rangey : array, list
-        range along vertical direction
-    start_x : float
-        starting value in horizontal direction.
-    start_y : float
-        starting value in vertical direction.
-    """
-    # try:
-    #     nx, ny = hdr['dimensions']
-    # except ValueError:
-    #     raise ValueError('Not a 2D flyscan (dimensions={})'
-    #                      ''.format(hdr['dimensions']))
-    nx, ny = dimv
-    # rangex, rangey = hdr['scan_range']
-    width = rangex[-1] - rangex[0]
-    height = rangey[-1] - rangey[0]
-
-    # macros = eval(hdr['subscan_0']['macros'], dict(array=np.array))
-    # start_x, start_y = macros['scan_starts']
-    dx = width / nx
-    dy = height / ny
-
-    #  original code use dx/2, dy/2, so value of the last column will not be interpolated
-    # grid_x = np.linspace(start_x, start_x + width + dx/2, nx)
-    # grid_y = np.linspace(start_y, start_y + height + dy/2, ny)
-    grid_x = np.linspace(start_x, start_x + width - dx / 2, nx)
-    grid_y = np.linspace(start_y, start_y + height - dy / 2, ny)
-
-    return grid_x, grid_y
-
-
-def interp1d_scan(dimv, rangex, rangey, start_x, start_y,
-                  x_data, y_data, spectrum, kind='linear',
-                  **kwargs):
-    """
-    Interpolate a 2D flyscan only over the fast-scanning direction
-    """
-
-    # grid_x, grid_y = fly2d_grid(hdr, x_data, y_data, plot=plot_points)
-    grid_x, grid_y = fly2d_grid(dimv, rangex, rangey, start_x, start_y, x_data, y_data)
-    # x_data = fly2d_reshape(hdr, x_data, verbose=False)
-    # grid_x = flip_data(grid_x)
-
-    if False:
-        mesh_x, mesh_y = np.meshgrid(grid_x, grid_y)
-        plt.figure()
-        if x_data is not None and y_data is not None:
-            plt.scatter(x_data, y_data, c='blue', label='actual')
-        plt.scatter(mesh_x, mesh_y, c='red', label='gridded',
-                    alpha=0.5)
-        plt.legend()
-        plt.show()
-
-    spectrum2 = np.zeros_like(spectrum)
-    for row in range(len(grid_y)):
-        spectrum2[row, :] = interp1d(x_data[row, :], spectrum[row, :],
-                                     kind=kind, bounds_error=False,
-                                     **kwargs)(grid_x)
-
-    return spectrum2
-
-
-def interp2d_scan(dimv, rangex, rangey, start_x, start_y,
-                  x_data, y_data, spectrum, kind='linear',
-                  **kwargs):
-    '''Interpolate a 2D flyscan over a grid, borrowed from Ken'''
-
-    new_x, new_y = fly2d_grid(dimv, rangex, rangey, start_x, start_y, x_data, y_data)
-
-    f = interp2d(x_data, y_data, spectrum, kind=kind, **kwargs)
-    return f(new_x, new_y)
