@@ -27,8 +27,7 @@ bound_options = ['none', 'lohi', 'fixed', 'lo', 'hi']
 fit_strategy_list = ['fit_with_tail', 'free_more',
                      'e_calibration', 'linear',
                      'adjust_element1', 'adjust_element2', 'adjust_element3']
-autofit_param = ['e_offset', 'e_linear', 'fwhm_offset', 'fwhm_fanoprime',
-                 'coherent_sct_energy']
+autofit_param = ['e_offset', 'e_linear', 'fwhm_offset', 'fwhm_fanoprime']
 
 
 class PreFitStatus(Atom):
@@ -183,7 +182,7 @@ class ElementController(object):
         remove_list = []
         non_element = ['compton', 'elastic', 'background']
         for k, v in six.iteritems(self.element_dict):
-            if v.norm <= threshv:
+            if v.norm < threshv:
                 remove_list.append(k)
         for name in remove_list:
             if name in non_element:
@@ -462,7 +461,17 @@ class GuessParamModel(Atom):
                     temp_dict[e] = ps
         self.EC.add_to_dict(temp_dict)
 
-    def manual_input(self):
+    def manual_input(self, userpeak_center=2.5):
+        """
+        Manually add an emission line (or peak).
+
+        Parameters
+        ----------
+        userpeak_center: float
+            Center of the user defined peak. Ignored if emission line other
+            than 'userpeak' is added
+        """
+
         default_area = 1e2
 
         # if self.e_name == 'escape':
@@ -482,9 +491,31 @@ class GuessParamModel(Atom):
         #     logger.info('{} peak is added'.format(self.e_name))
         #
         # else:
+
+        # Add the new data entry to the parameter dictionary. This operation is necessary for 'userpeak'
+        #   lines, because they need to be placed to the specific position (by setting 'delta_center'
+        #   parameter, while regular element lines are placed to their default positions.
+        d_energy = userpeak_center - 5.0
+
+        # PC.params will contain a deepcopy of 'self.param_new' with the new line added
+        PC = ParamController(self.param_new, [self.e_name])
+
+        if "userpeak" in self.e_name.lower():
+            # Default values for 'delta_center'
+            dc = copy.deepcopy(PC.params[f"{self.e_name}_delta_center"])
+            # Modify the default values in the dictionary of parameters
+            PC.params[f"{self.e_name}_delta_center"]["value"] = d_energy
+            PC.params[f"{self.e_name}_delta_center"]["min"] = d_energy - (dc["value"] - dc["min"])
+            PC.params[f"{self.e_name}_delta_center"]["max"] = d_energy + (dc["max"] - dc["value"])
+
+        param_tmp = PC.params
+
+        # 'self.param_new' is used to provide 'hint' values for the model, but all active
+        #    emission lines in 'elemental_lines' will be included in the model.
+        #  The model will contain lines in 'elemental_lines', Compton and elastic
         x, data_out, area_dict = calculate_profile(self.x0,
                                                    self.y0,
-                                                   self.param_new,
+                                                   param_tmp,
                                                    elemental_lines=[self.e_name],
                                                    default_area=default_area)
 
@@ -495,6 +526,10 @@ class GuessParamModel(Atom):
         if self.e_name not in data_out:
             raise Exception(f"Failed to add the emission line '{self.e_name}': line is not activated.")
 
+        # If model was generated successfully (the emission line was successfully added), then
+        #   make temporary parameters permanent
+        self.param_new = param_tmp
+
         ratio_v = self.add_element_intensity / np.max(data_out[self.e_name])
 
         ps = PreFitStatus(z=get_Z(self.e_name),
@@ -564,24 +599,29 @@ class GuessParamModel(Atom):
         # Create the list of selected emission lines such as Ca_K, K_K, etc.
         #   No pileup or user peaks
         pure_peak_list = [n for n in self.result_dict_names if n in peak_list]
-        self.n_selected_elines_for_fitting = len(peak_list)
+        self.n_selected_elines_for_fitting = len(self.result_dict_names)
         self.n_selected_pure_elines_for_fitting = len(pure_peak_list)
 
-        # logger.info('The full list for fitting is {}'.format(self.result_dict_names))
+        logger.info(f"The full list for fitting is {self.result_dict_names}")
 
-    def find_peak(self, threshv=0.1):
+    def find_peak(self, *, threshv=0.1, elemental_lines=None):
         """
         Run automatic peak finding, and save results as dict of object.
 
         Parameters
         ----------
-        threshv : float
+        threshv: float
             The value will not be shown on GUI if it is smaller than the threshold.
+
+        elemental_lines: list(str)
+            The list of elemental lines to find. If ``None``, then all supported
+            lines (K, L and M) are searched.
         """
         self.define_range()  # in case the energy calibraiton changes
         self.prefit_x, out_dict, area_dict = linear_spectrum_fitting(self.x0,
                                                                      self.y0,
-                                                                     self.param_new)
+                                                                     self.param_new,
+                                                                     elemental_lines=elemental_lines)
         logger.info('Energy range: {}, {}'.format(
             self.param_new['non_fitting_values']['energy_bound_low']['value'],
             self.param_new['non_fitting_values']['energy_bound_high']['value']))
