@@ -8,6 +8,7 @@ from matplotlib.figure import Figure
 from matplotlib.axes import Axes
 import matplotlib.pyplot as plt
 from matplotlib.lines import Line2D
+from matplotlib.collections import BrokenBarHCollection
 from collections import OrderedDict
 
 from atom.api import Atom, Str, observe, Typed, Int, List, Dict, Float, Bool
@@ -108,6 +109,10 @@ class LinePlotModel(Atom):
     plot_exp_obj = Typed(Line2D)
     show_exp_opt = Bool(False)
 
+    # Reference to artist responsible for displaying the selected range of energies on the plot
+    plot_energy_barh = Typed(BrokenBarHCollection)
+    t_bar = Typed(object)
+
     plot_exp_list = List()
     data_sets = Typed(OrderedDict)
 
@@ -149,6 +154,13 @@ class LinePlotModel(Atom):
     # Reference to GuessParamModel object
     param_model = Typed(object)
 
+    # Location of the vertical (mouse-selected) marker on the plot.
+    # Value is in kev. Negative value - no marker is placed.
+    vertical_marker_kev = Float(-1)
+    # Reference to the respective Matplotlib artist
+    line_vertical_marker = Typed(object)
+    vertical_marker_is_visible = Bool(False)
+
     def __init__(self, param_model):
 
         # Reference to GuessParamModel object
@@ -157,6 +169,7 @@ class LinePlotModel(Atom):
         self.data = None
 
         self._fig = plt.figure(figsize=(3, 2))
+
         self._ax = self._fig.add_subplot(111)
         try:
             self._ax.set_axis_bgcolor('lightgrey')
@@ -218,7 +231,19 @@ class LinePlotModel(Atom):
         # Reset currently selected element_id (mostly to reset GUI elements)
         self.element_id = 0
 
+    def init_mouse_event(self):
+        """Set up callback for mouse button-press event"""
+        # Reference to the toolbar
+        self.t_bar = self._fig.canvas.toolbar
+        # Set callback for Button Press event
+        self._fig.canvas.mpl_connect("button_press_event", self.canvas_onpress)
+
     def _update_canvas(self):
+        # It may be sufficient to initialize the event only once, but at this point
+        #   it seems to be the most reliable option. May be changed in the future.
+        self.init_mouse_event()
+        self.plot_vertical_marker()
+
         self._ax.legend(loc=2)
         try:
             self._ax.legend(framealpha=0.2).set_draggable(True)
@@ -264,11 +289,72 @@ class LinePlotModel(Atom):
             return
         self.incident_energy = self.parameters['coherent_sct_energy']['value']
 
+    def set_incident_energy(self, change):
+        """
+        The observer function that changes the value of incident energy
+        and upper bound for fitted energy range. Should not be called directly.
+
+        Parameters
+        ----------
+
+        change : dict
+            ``change["value"]`` is the new value of incident energy
+        """
+        self.change_incident_energy(change["value"])
+
+    def change_incident_energy(self, energy_new):
+        """
+        The function that perfroms the changes the value of incident energy
+        and upper bound for fitted energy range.
+
+        Parameters
+        ----------
+
+        incident_energy : float
+            New value of incident energy
+        """
+
+        margin = 0.8  # Value by which the upper bound of the range used for fitting
+        #               exceeds the incident energy. Selected for convenience, but
+        #               is subject to change. This is the place to change it to take effect
+        #               throughout the program.
+
+        # Limit the number of decimal points for better visual presentation
+        energy_new = round(energy_new, ndigits=6)
+        # Change the value twice to ensure that all observer functions are called
+        self.incident_energy = energy_new + 1.0  # Arbitrary number different from 'energy_new'
+        self.incident_energy = energy_new
+        if 'coherent_sct_energy' in self.param_model.param_new:
+            self.param_model.param_new['coherent_sct_energy']['value'] = energy_new
+
+        # Change the value twice to ensure that all observer functions are called
+        self.param_model.energy_bound_high_buf = energy_new + 1.8  # Arbitrary number
+        upper_bound = energy_new + margin
+        # Limit the number of decimal points for better visual presentation
+        upper_bound = round(upper_bound, ndigits=5)
+        self.param_model.energy_bound_high_buf = upper_bound
+
     @observe('scale_opt')
     def _new_opt(self, change):
         self.log_linear_plot()
         self._update_canvas()
 
+    def energy_bound_high_update(self, change):
+        """Observer function for 'param_model.energy_bound_high_buf'"""
+        if self.data is None:
+            return
+        self.plot_selected_energy_range(e_high=change["value"])
+        self.plot_vertical_marker(e_high=change["value"])
+        self._update_canvas()
+
+    def energy_bound_low_update(self, change):
+        """Observer function for 'param_model.energy_bound_low_buf'"""
+        if self.data is None:
+            return
+        self.plot_selected_energy_range(e_low=change["value"])
+        self.plot_vertical_marker(e_low=change["value"])
+        self._update_canvas()
+
     def log_linear_plot(self):
         if self.plot_type[self.scale_opt] == 'LinLog':
             self._ax.set_yscale('log')
@@ -328,6 +414,8 @@ class LinePlotModel(Atom):
 
         self._set_eline_select_controls()
 
+        self.plot_selected_energy_range()
+
         # _show_hide_exp_plot is called to show or hide current plot based
         #           on the state of _show_exp_opt flag
         self._show_hide_exp_plot(self.show_exp_opt or self.plot_exp_opt)
@@ -414,6 +502,87 @@ class LinePlotModel(Atom):
         data_arr = np.asarray(self.data)
         self.exp_data_update({'value': data_arr})
 
+    def plot_vertical_marker(self, *, e_low=None, e_high=None):
+
+        self._vertical_marker_set_inside_range(e_low=e_low, e_high=e_high)
+
+        x_v = (self.vertical_marker_kev, self.vertical_marker_kev)
+        y_v = (-1e30, 1e30)  # This will cover the range of possible values of accumulated counts
+
+        if self.line_vertical_marker:
+            self._ax.lines.remove(self.line_vertical_marker)
+            self.line_vertical_marker = None
+        if self.vertical_marker_is_visible:
+            self.line_vertical_marker, = self._ax.plot(x_v, y_v, color="blue")
+
+    def set_plot_vertical_marker(self, marker_position=None):
+        """
+        The function is called when setting the position of the marker interactively
+
+        If the parameter `marker_position` is `None`, then don't set or change the value.
+        Just make the marker visible.
+        """
+
+        # Ignore the new value if it is outside the range of selected energies
+        if marker_position is not None:
+            e_low = self.param_model.param_new['non_fitting_values']['energy_bound_low']['value']
+            e_high = self.param_model.param_new['non_fitting_values']['energy_bound_high']['value']
+            if (marker_position >= e_low) and (marker_position <= e_high):
+                self.vertical_marker_kev = marker_position
+
+        # Compute peak intensity. The displayed value will change only for user defined peak,
+        #   since it is moved to the position of the marker.
+        self.compute_manual_peak_intensity()
+
+        # Make the marker visible
+        self.vertical_marker_is_visible = True
+
+        # Update the location of the marker and the canvas
+        self.plot_vertical_marker()
+        self._update_canvas()
+
+    def hide_plot_vertical_marker(self):
+        """Hide vertical marker"""
+        self.vertical_marker_is_visible = False
+        self.plot_vertical_marker()
+        self._update_canvas()
+
+    def plot_selected_energy_range(self, *, e_low=None, e_high=None):
+        """
+        Plot the range of energies selected for processing. The range may be optionally
+        provided as arguments. The range values that are not provided, are read from
+        globally accessible dictionary of parameters. The values passed as arguments
+        are mainly used if the function is called during interactive update of the
+        range, when the order of update is undetermined and the parameter dictionary
+        may be updated after the function is called.
+        """
+        # The range of energy selected for analysis
+        if e_low is None:
+            e_low = self.param_model.param_new['non_fitting_values']['energy_bound_low']['value']
+        if e_high is None:
+            e_high = self.param_model.param_new['non_fitting_values']['energy_bound_high']['value']
+
+        n_x = 4096  # Set to the maximum possible number of points
+
+        # Generate the values for 'energy' axis
+        x_v = (self.parameters['e_offset']['value'] +
+               np.arange(n_x) *
+               self.parameters['e_linear']['value'] +
+               np.arange(n_x) ** 2 *
+               self.parameters['e_quadratic']['value'])
+
+        ss = (x_v < e_high) & (x_v > e_low)
+        y_min, y_max = 1e-30, 1e30  # Select the max and min values for plotted rectangles
+
+        # Remove the plot if it exists
+        if self.plot_energy_barh in self._ax.collections:
+            self._ax.collections.remove(self.plot_energy_barh)
+
+        # Create the new plot (based on new parameters if necessary
+        self.plot_energy_barh = BrokenBarHCollection.span_where(
+            x_v, ymin=y_min, ymax=y_max, where=ss, facecolor='white', edgecolor='yellow', alpha=1)
+        self._ax.add_collection(self.plot_energy_barh)
+
     def plot_multi_exp_data(self):
         while(len(self.plot_exp_list)):
             self.plot_exp_list.pop().remove()
@@ -422,7 +591,7 @@ class LinePlotModel(Atom):
 
         self.max_v = 1.0
         m = 0
-        for (k, v) in six.iteritems(self.data_sets):
+        for (k, v) in self.data_sets.items():
             if v.plot_index:
 
                 data_arr = np.asarray(v.data)
@@ -443,6 +612,8 @@ class LinePlotModel(Atom):
                 self.plot_exp_list.append(plot_exp_obj)
                 m += 1
 
+        self.plot_selected_energy_range()
+
         self._update_ylimit()
         self.log_linear_plot()
         self._update_canvas()
@@ -636,6 +807,68 @@ class LinePlotModel(Atom):
             ename = None
         return ename
 
+    def _vertical_marker_set_inside_range(self, *, e_low=None, e_high=None):
+        """
+        Don't move the marker if it is inside range. If it is outside range,
+        then set the marker to the center of the range
+        """
+        # The range of energy selected for analysis
+        if e_low is None:
+            e_low = self.param_model.param_new['non_fitting_values']['energy_bound_low']['value']
+        if e_high is None:
+            e_high = self.param_model.param_new['non_fitting_values']['energy_bound_high']['value']
+
+        # By default, place the marker in the middle of the range if its original position
+        #   is outside the range
+        if (self.vertical_marker_kev > e_high) or (self.vertical_marker_kev < e_low):
+            self.vertical_marker_kev = (e_low + e_high) / 2.0
+
+    def _fill_elist(self):
+
+        _elist = []
+
+        incident_energy = self.incident_energy
+        k_len = len(K_TRANSITIONS)
+        l_len = len(L_TRANSITIONS)
+        m_len = len(M_TRANSITIONS)
+
+        ename = self.get_element_line_name_by_id(self.element_id, include_user_peaks=True)
+
+        if ename is not None:
+
+            _elist = []
+            if ename.lower().startswith("userpeak"):
+                # Make sure that the marker is in the selected range of energies
+                self._vertical_marker_set_inside_range()
+                # The tuple structure: (center_energy, ratio)
+                _elist.append((self.vertical_marker_kev, 1.0))
+
+            elif '_K' in ename:
+                e = Element(ename[:-2])
+                if e.cs(incident_energy)['ka1'] != 0:
+                    for i in range(k_len):
+                        _elist.append((e.emission_line.all[i][1],
+                                       e.cs(incident_energy).all[i][1]
+                                       / e.cs(incident_energy).all[0][1]))
+
+            elif '_L' in ename:
+                e = Element(ename[:-2])
+                if e.cs(incident_energy)['la1'] != 0:
+                    for i in range(k_len, k_len+l_len):
+                        _elist.append((e.emission_line.all[i][1],
+                                       e.cs(incident_energy).all[i][1]
+                                       / e.cs(incident_energy).all[k_len][1]))
+
+            else:
+                e = Element(ename[:-2])
+                if e.cs(incident_energy)['ma1'] != 0:
+                    for i in range(k_len+l_len, k_len+l_len+m_len):
+                        _elist.append((e.emission_line.all[i][1],
+                                       e.cs(incident_energy).all[i][1]
+                                       / e.cs(incident_energy).all[k_len+l_len][1]))
+
+            return _elist
+
     @observe('element_id')
     def set_element(self, change):
 
@@ -653,43 +886,19 @@ class LinePlotModel(Atom):
             return
 
         incident_energy = self.incident_energy
-        k_len = len(K_TRANSITIONS)
-        l_len = len(L_TRANSITIONS)
-        m_len = len(M_TRANSITIONS)
-
-        ename = self.get_element_line_name_by_id(self.element_id)
+        ename = self.get_element_line_name_by_id(self.element_id, include_user_peaks=True)
 
         if ename is not None:
 
             logger.debug('Plot emission line for element: '
                          '{} with incident energy {}'.format(self.element_id,
                                                              incident_energy))
-            self.elist = []
-
-            if '_K' in ename:
-                e = Element(ename[:-2])
-                if e.cs(incident_energy)['ka1'] != 0:
-                    for i in range(k_len):
-                        self.elist.append((e.emission_line.all[i][1],
-                                           e.cs(incident_energy).all[i][1]
-                                           / e.cs(incident_energy).all[0][1]))
-
-            elif '_L' in ename:
-                e = Element(ename[:-2])
-                if e.cs(incident_energy)['la1'] != 0:
-                    for i in range(k_len, k_len+l_len):
-                        self.elist.append((e.emission_line.all[i][1],
-                                           e.cs(incident_energy).all[i][1]
-                                           / e.cs(incident_energy).all[k_len][1]))
 
+            _elist = self._fill_elist()
+            if not ename.lower().startswith("userpeak"):
+                self.elist = _elist
             else:
-                e = Element(ename[:-2])
-                if e.cs(incident_energy)['ma1'] != 0:
-                    for i in range(k_len+l_len, k_len+l_len+m_len):
-                        self.elist.append((e.emission_line.all[i][1],
-                                           e.cs(incident_energy).all[i][1]
-                                           / e.cs(incident_energy).all[k_len+l_len][1]))
-
+                self.elist = []
             self.plot_emission_line()
             self._update_canvas()
 
@@ -752,11 +961,13 @@ class LinePlotModel(Atom):
 
     def add_peak_manual(self):
 
-        self.param_model.manual_input()
+        self.param_model.manual_input(userpeak_center=self.vertical_marker_kev)
         self.param_model.EC.update_peak_ratio()
         self.param_model.update_name_list()
         self.param_model.data_for_plot()
 
+        self.hide_plot_vertical_marker()
+
         self.plot_fit(self.param_model.prefit_x,
                       self.param_model.total_y,
                       self.param_model.auto_fit_all)
@@ -792,12 +1003,70 @@ class LinePlotModel(Atom):
         self.show_fit_opt = True
         self.param_model.update_name_list()
 
+    def _compute_intensity(self, elist):
+        # Some default value
+        intensity = 1000.0
+
+        if self.data is not None and self.parameters is not None \
+                and self.param_model.prefit_x is not None \
+                and len(self.data) > 1 and len(self.param_model.prefit_x) > 1:
+
+            # Range of energies in fitting results
+            e_fit_min = self.param_model.prefit_x[0]
+            e_fit_max = self.param_model.prefit_x[-1]
+            de_fit = (e_fit_max - e_fit_min) / (len(self.param_model.prefit_x) - 1)
+
+            e_raw_min = self.parameters['e_offset']['value']
+            e_raw_max = self.parameters['e_offset']['value'] + \
+                (len(self.data) - 1) * self.parameters['e_linear']['value'] + \
+                (len(self.data) - 1) ** 2 * self.parameters['e_quadratic']['value']
+            de_raw = (e_raw_max - e_raw_min) / (len(self.data) - 1)
+
+            # Note: the above algorithm for finding 'de_raw' is far from perfect but will
+            #    work for now. As a result 'de_fit' and 'de_raw' == self.parameters['e_linear']['value'].
+            #    So the quadratic coefficent is ignored. This is OK, since currently
+            #    quadratic coefficient is always ZERO. When the program is rewritten,
+            #    the complete algorithm should be revised.
+
+            # Find the line with maximum energy. It must come first in the list,
+            #    but let's check just to make sure
+            max_line_energy, max_line_intensity = 0, 0
+            if elist:
+                for e, i in elist:
+                    # e - line peak energy
+                    # i - peak intensity relative to maximum peak
+                    if e >= e_fit_min and e <= e_fit_max and e > e_raw_min and e < e_raw_max:
+                        if max_line_intensity < i:
+                            max_line_energy, max_line_intensity = e, i
+
+            # Find the index of peak maximum in the 'fitted' data array
+            n = (max_line_energy - e_fit_min) / de_fit
+            n = np.clip(n, 0, len(self.param_model.total_y) - 1)
+            n_fit = int(round(n))
+            # Find the index of peak maximum in the 'raw' data array
+            n = (max_line_energy - e_raw_min) / de_raw
+            n = np.clip(n, 0, len(self.data) - 1)
+            n_raw = int(round(n))
+            # Intensity of the fitted data at the peak
+            in_fit = self.param_model.total_y[n_fit]
+            # Intensity of the raw data at the peak
+            in_raw = self.data[n_raw]
+            # The estimated peak intensity is the difference:
+            intensity = in_raw - in_fit
+
+            # The following step is questionable. We assign some reasonably small number.
+            #   The desired value can always be manually entered
+            if intensity < 0.0:
+                intensity = abs(in_raw / 100)
+
+        return intensity
+
     def compute_manual_peak_intensity(self, n_id=None):
 
         if n_id is None:
             n_id = self.element_id
 
-        if not self.is_element_line_id_valid(n_id):
+        if not self.is_element_line_id_valid(n_id, include_user_peaks=True):
             # This is typicall the case when n_id==0
             intensity = 0.0
             if self.is_element_line_id_valid(n_id, include_user_peaks=True):
@@ -807,70 +1076,15 @@ class LinePlotModel(Atom):
                     intensity = self.param_model.EC.element_dict[name].maxv
 
         else:
-            if self.is_line_in_selected_list(n_id):
-                name = self.get_element_line_name_by_id(n_id)
+            if self.is_line_in_selected_list(n_id, include_user_peaks=True):
+                name = self.get_element_line_name_by_id(n_id, include_user_peaks=True)
                 intensity = self.param_model.EC.element_dict[name].maxv
-
             else:
+                _elist = self._fill_elist()
+                intensity = self._compute_intensity(_elist)
 
-                # Some default value
-                intensity = 1000.0
-
-                if self.data is not None and self.parameters is not None \
-                        and self.param_model.prefit_x is not None \
-                        and len(self.data) > 1 and len(self.param_model.prefit_x) > 1:
-
-                    # Range of energies in fitting results
-                    e_fit_min = self.param_model.prefit_x[0]
-                    e_fit_max = self.param_model.prefit_x[-1]
-                    de_fit = (e_fit_max - e_fit_min) / (len(self.param_model.prefit_x) - 1)
-
-                    e_raw_min = self.parameters['e_offset']['value']
-                    e_raw_max = self.parameters['e_offset']['value'] + \
-                        (len(self.data) - 1) * self.parameters['e_linear']['value'] + \
-                        (len(self.data) - 1) ** 2 * self.parameters['e_quadratic']['value']
-                    de_raw = (e_raw_max - e_raw_min) / (len(self.data) - 1)
-
-                    # Note: the above algorithm for finding 'de_raw' is far from perfect but will
-                    #    work for now. As a result 'de_fit' and 'de_raw' == self.parameters['e_linear']['value'].
-                    #    So the quadratic coefficent is ignored. This is OK, since currently
-                    #    quadratic coefficient is always ZERO. When the program is rewritten,
-                    #    the complete algorithm should be revised.
-
-                    # Find the line with maximum energy. It must come first in the list,
-                    #    but let's check just to make sure
-                    max_line_energy, max_line_intensity = 0, 0
-                    if self.elist:
-                        for e, i in self.elist:
-                            # e - line peak energy
-                            # i - peak intensity relative to maximum peak
-                            if e >= e_fit_min and e <= e_fit_max and e > e_raw_min and e < e_raw_max:
-                                if max_line_intensity < i:
-                                    max_line_energy, max_line_intensity = e, i
-
-                    # Find the index of peak maximum in the 'fitted' data array
-                    n = (max_line_energy - e_fit_min) / de_fit
-                    n = np.clip(n, 0, len(self.param_model.total_y) - 1)
-                    n_fit = int(round(n))
-                    # Find the index of peak maximum in the 'raw' data array
-                    n = (max_line_energy - e_raw_min) / de_raw
-                    n = np.clip(n, 0, len(self.data) - 1)
-                    n_raw = int(round(n))
-                    # Intensity of the fitted data at the peak
-                    in_fit = self.param_model.total_y[n_fit]
-                    # Intensity of the raw data at the peak
-                    in_raw = self.data[n_raw]
-                    # The estimated peak intensity is the difference:
-                    intensity = in_raw - in_fit
-
-                    # The following step is questionable. We assign some reasonably small number.
-                    #   The desired value can always be manually entered
-                    if intensity < 0.0:
-                        intensity = abs(in_raw / 100)
-
-        # Round for nicer printing
-        intensity = float(f"{intensity:.2f}")
-        self.param_model.add_element_intensity = intensity
+        # Round the intensity for nicer printing
+        self.param_model.add_element_intensity = round(intensity, 2)
 
     # def plot_autofit(self):
     #     sum_y = 0
@@ -1142,3 +1356,13 @@ class LinePlotModel(Atom):
                 pass
 
         # self._update_canvas()
+
+    def canvas_onpress(self, event):
+        """Callback, mouse button pressed"""
+        if (self.t_bar.mode == ""):
+            if event.inaxes == self._ax:
+                if event.button == 1:
+                    xd = event.xdata
+                    self.set_plot_vertical_marker(marker_position=xd)
+                else:
+                    self.hide_plot_vertical_marker()
