@@ -9,10 +9,17 @@ import json
 import multiprocessing
 import pandas as pd
 import math
+import time as ttime
 
 import logging
 import warnings
 
+from ..core.utils import convert_time_to_nexus_string
+from .scan_metadata import ScanMetadataXRF
+
+import pyxrf
+pyxrf_version = pyxrf.__version__
+
 logger = logging.getLogger()
 warnings.filterwarnings('ignore')
 
@@ -182,12 +189,12 @@ def fetch_data_from_db(runid, fpath=None,
     return data
 
 
-def make_hdf(start, end=None, *, fname=None,
+def make_hdf(start, end=None, *, fname=None, wd=None,
              fname_add_version=False,
              completed_scans_only=False,
              file_overwrite_existing=False,
              prefix='scan2D_',
-             create_each_det=True, save_scaler=True,
+             create_each_det=False, save_scaler=True,
              num_end_lines_excluded=None):
     """
     Load data from database and save it in HDF5 files.
@@ -228,6 +235,10 @@ def make_hdf(start, end=None, *, fname=None,
     fname : string, optional keyword parameter
         path to save data file when ``end`` is ``None`` (only one scan is processed).
         File name is created automatically if ``fname`` is not specified.
+    wd : str
+        working directory, the file(s) will be created in this directory. The directory
+        will be created if it does not exist. If ``wd`` is not specified, then the file(s)
+        will be saved to the current directory.
     fname_add_version : bool, keyword parameter
         True: if file already exists, then file version is added to the file name
         so that it becomes unique in the current directory. The version is
@@ -290,11 +301,19 @@ def make_hdf(start, end=None, *, fname=None,
         The number of lines at the end of the scan that will not be saved to the data file.
     """
 
+    if wd:
+        # Create the directory
+        wd = os.path.expanduser(wd)
+        wd = os.path.abspath(wd)  # 'make_dirs' does not accept paths that contain '..'
+        os.makedirs(wd, exist_ok=True)  # Does nothing if the directory already exists
+
     if end is None:
         # Load one scan with ID specified by ``start``
         #   If there is a problem while reading the scan, the exception is raised.
         if fname is None:
             fname = prefix+str(start)+'.h5'
+            if wd:
+                fname = os.path.join(wd, fname)
         fetch_data_from_db(start, fpath=fname,
                            create_each_det=create_each_det,
                            fname_add_version=fname_add_version,
@@ -309,9 +328,11 @@ def make_hdf(start, end=None, *, fname=None,
         #   then the scan is skipped and the next scan is processed
         datalist = range(start, end+1)
         for v in datalist:
-            filename = prefix+str(v)+'.h5'
+            fname = prefix+str(v)+'.h5'
+            if wd:
+                fname = os.path.join(wd, fname)
             try:
-                fetch_data_from_db(v, fpath=filename,
+                fetch_data_from_db(v, fpath=fname,
                                    create_each_det=create_each_det,
                                    fname_add_version=fname_add_version,
                                    completed_scans_only=completed_scans_only,
@@ -347,6 +368,128 @@ def _is_scan_complete(hdr):
     return bool(hdr.stop)
 
 
+def _extract_metadata_from_header(hdr):
+    """
+    Extract metadata from start and stop document. Metadata extracted from other document
+    in the scan are beamline specific and added to dictionary at later time.
+    """
+
+    start_document = hdr.start
+
+    mdata = ScanMetadataXRF()
+
+    data_locations = {
+        "scan_id": ["scan_id"],
+        "scan_uid": ["uid"],
+        "scan_instrument_id": ["beamline_id"],
+        "scan_instrument_name": [],
+        "scan_time_start": ["time"],
+        "scan_time_start_utc": ["time"],
+
+        "instrument_mono_incident_energy": ["beamline_status/energy"],
+        "instrument_beam_current": [],
+        "instrument_detectors": ["detectors"],
+
+        "sample_name": ["sample/name", "sample"],
+
+        "experiment_plan_name": ["plan_name"],
+        "experiment_plan_type": ["plan_type"],
+
+        "proposal_num": ["proposal/proposal_num"],
+        "proposal_title": ["proposal/proposal_title"],
+        "proposal_PI_lastname": ["proposal/PI_lastname"],
+        "proposal_saf_num": ["proposal/saf_num"],
+        "proposal_cycle": ["proposal/cycle"]
+    }
+
+    for key, locations in data_locations.items():
+        # Go to the next key if no location is defined for the current key.
+        #   No locations means that the data is not yet defined in start document on any beamline
+        #   Multiple locations point to locations at different beamlines
+        if not locations:
+            continue
+
+        # For each metadata key there could be none, one or multiple locations in the start document
+        for loc in locations:
+            path = loc.split('/')  #
+            ref = start_document
+            for n, p in enumerate(path):
+                if n >= len(path) - 1:
+                    break
+                # 'ref' must always point to dictionary
+                if not isinstance(ref, dict):
+                    ref = None
+                    break
+                if p in ref:
+                    ref = ref[p]
+                else:
+                    ref = None
+                    break
+            # At this point 'ref' must be a dictionary
+            value = None
+            if ref is not None and isinstance(ref, dict):
+                if path[-1] in ref:
+                    value = ref[path[-1]]
+            # Now we finally arrived to the end of the path: the 'value' must be a scalar or a list
+            if value is not None and not isinstance(value, dict):
+                if path[-1] == 'time':
+                    if key.endswith("_utc"):
+                        value = convert_time_to_nexus_string(ttime.gmtime(value))
+                    else:
+                        value = convert_time_to_nexus_string(ttime.localtime(value))
+                mdata[key] = value
+                break
+
+    stop_document = hdr.stop
+
+    if stop_document:
+
+        if "time" in stop_document:
+            t = stop_document["time"]
+            mdata["scan_time_stop"] = convert_time_to_nexus_string(ttime.localtime(t))
+            mdata["scan_time_stop_utc"] = convert_time_to_nexus_string(ttime.gmtime(t))
+
+        if "exit_status" in stop_document:
+            mdata["scan_exit_status"] = stop_document["exit_status"]
+
+    else:
+
+        mdata["scan_exit_status"] = "incomplete"
+
+    # Add full beamline name (if available, otherwise don't create the entry).
+    #   Also, don't overwrite the existing name if it was read from the start document
+    if "scan_instrument_id" in mdata and "scan_instrument_name" not in mdata:
+        instruments = {
+            "srx": "Submicron Resolution X-ray Spectroscopy",
+            "hxn": "Hard X-ray Nanoprobe",
+            "tes": "Tender Energy X-ray Absorption Spectroscopy",
+            "xfm": "X-ray Fluorescence Microprobe"
+        }
+        iname = instruments.get(mdata["scan_instrument_id"].lower(), "")
+        if iname:
+            mdata["scan_instrument_name"] = iname
+
+    return mdata
+
+
+def _get_metadata_from_descriptor_document(hdr, *, data_key, stream_name='baseline'):
+
+    # Returns None if the parameter is not found
+
+    value = None
+    docs = hdr.documents(stream_name=stream_name)
+    for name, doc in docs:
+        if (name != "event") or ("descriptor" not in doc):
+            continue
+        try:
+            value = doc['data'][data_key]
+            break  # Don't go through the rest of the documents
+        except Exception:
+            pass
+
+    return value
+
+
 def map_data2D_hxn(runid, fpath,
                    create_each_det=False,
                    fname_add_version=False,
@@ -391,6 +534,7 @@ def map_data2D_hxn(runid, fpath,
         save data to hdf5 file if True
     """
     hdr = db[runid]
+    runid = hdr.start["scan_id"]  # Replace with the true value (runid may be relative, such as -2)
 
     if completed_scans_only and not _is_scan_complete(hdr):
         raise Exception("Scan is incomplete. Only completed scans are currently processed.")
@@ -403,6 +547,20 @@ def map_data2D_hxn(runid, fpath,
     data_output = []
 
     start_doc = hdr['start']
+    # The dictionary holding scan metadata
+    mdata = _extract_metadata_from_header(hdr)
+    # Some metadata is located at specific places in the descriptor documents
+    # Search through the descriptor documents for the metadata
+    v = _get_metadata_from_descriptor_document(hdr, data_key="beamline_status_beam_current",
+                                               stream_name="baseline")
+    if v is not None:
+        mdata["instrument_beam_current"] = v
+
+    v = _get_metadata_from_descriptor_document(hdr, data_key="energy",
+                                               stream_name="baseline")
+    if v is not None:
+        mdata["instrument_mono_incident_energy"] = v
+
     if 'dimensions' in start_doc:
         datashape = start_doc.dimensions
     elif 'shape' in start_doc:
@@ -442,20 +600,20 @@ def map_data2D_hxn(runid, fpath,
                           det_list=det_list,
                           pos_list=pos_list,
                           scaler_list=scaler_list,
+                          create_each_det=create_each_det,
                           fly_type=fly_type, subscan_dims=subscan_dims,
                           spectrum_len=4096)
     if output_to_file:
         # output to file
         print('Saving data to hdf file.')
-        fpath = write_db_to_hdf_base(fpath, data_out,
+        fpath = write_db_to_hdf_base(fpath, data_out, metadata=mdata,
                                      fname_add_version=fname_add_version,
                                      file_overwrite_existing=file_overwrite_existing,
                                      create_each_det=create_each_det)
 
     detector_name = "xpress3"
-    d_dict = {"dataset": data_out, "file_name": fpath, "detector_name": detector_name}
+    d_dict = {"dataset": data_out, "file_name": fpath, "detector_name": detector_name, "metadata": mdata}
     data_output.append(d_dict)
-
     return data_output
 
     # write_db_to_hdf(fpath, data, datashape,
@@ -547,12 +705,15 @@ def map_data2D_srx(runid, fpath,
     dict of data in 2D format matching x,y scanning positions
     """
     hdr = db[runid]
+    runid = hdr.start["scan_id"]  # Replace with the true value (runid may be relative, such as -2)
 
     if completed_scans_only and not _is_scan_complete(hdr):
         raise Exception("Scan is incomplete. Only completed scans are currently processed.")
 
     spectrum_len = 4096
     start_doc = hdr['start']
+    # The dictionary holding scan metadata
+    mdata = _extract_metadata_from_header(hdr)
     plan_n = start_doc.get('plan_name')
 
     # Load configuration file
@@ -579,10 +740,21 @@ def map_data2D_srx(runid, fpath,
         fly_type = None
 
         if num_end_lines_excluded is None:
-            # vertical first then horizontal, assuming fast scan on x
-            datashape = [start_doc['shape'][1], start_doc['shape'][0]]
+            # It seems like the 'shape' in plan is in the form of [y, x], where
+            #    y - is the vertical and x is horizontal axis. This matches the
+            #    shape of the matrix that is used for storage of the maps.
+            #    In step scan, the results are represented as 1D array, not 2D array,
+            #    so it needs to be reshaped before processing. So the datashape
+            #    needs to be determined correctly.
+            # We also assume that scanning is performed along the x-axis first
+            #    before stepping along y-axis. Snaking may be on or off.
+            #    Different order (along y-axis first, then along x-axis) will require
+            #    some additional parameter in the start document to indicate this.
+            #    And the 'datashape' will need to be set the opposite way. Also
+            #    the map representation will be transposed.
+            datashape = [start_doc['shape'][0], start_doc['shape'][1]]
         else:
-            datashape = [start_doc['shape'][1] - num_end_lines_excluded, start_doc['shape'][0]]
+            datashape = [start_doc['shape'][0] - num_end_lines_excluded, start_doc['shape'][1]]
 
         snake_scan = start_doc.get('snaking')
         if snake_scan[1] is True:
@@ -622,11 +794,12 @@ def map_data2D_srx(runid, fpath,
                     fly_type=fly_type,
                     base_val=config_data['base_value'])  # base value shift for ic
                 fpath_out = write_db_to_hdf_base(
-                    fpath_out, data_out,
+                    fpath_out, data_out, metadata=mdata,
                     fname_add_version=fname_add_version,
                     file_overwrite_existing=file_overwrite_existing,
                     create_each_det=create_each_det)
-                d_dict = {"dataset": data_out, "file_name": fpath_out, "detector_name": "xs"}
+                d_dict = {"dataset": data_out, "file_name": fpath_out,
+                          "detector_name": "xs", "metadata": mdata}
                 data_output.append(d_dict)
 
             if 'xs2' in hdr.start.detectors:
@@ -643,11 +816,12 @@ def map_data2D_srx(runid, fpath,
                     fly_type=fly_type,
                     base_val=config_data['base_value'])  # base value shift for ic
                 fpath_out = write_db_to_hdf_base(
-                    fpath_out, data_out,
+                    fpath_out, data_out, metadata=mdata,
                     fname_add_version=fname_add_version,
                     file_overwrite_existing=file_overwrite_existing,
                     create_each_det=create_each_det)
-                d_dict = {"dataset": data_out, "file_name": fpath_out, "detector_name": "xs"}
+                d_dict = {"dataset": data_out, "file_name": fpath_out,
+                          "detector_name": "xs", "metadata": mdata}
                 data_output.append(d_dict)
 
         return data_output
@@ -729,55 +903,62 @@ def map_data2D_srx(runid, fpath,
 
             detector_field_exists = True
 
-            for m, v in enumerate(e):
+            # This 'try' block was added in response to the request to retrieve data after
+            #   detector failure (empty files were saved by Xpress3). The program is supposed
+            #   to retrieve 'good' data from the scan.
+            try:
+                for m, v in enumerate(e):
 
-                if m == 0:
+                    if m == 0:
 
-                    # Check if detector field does not exist. If not, then the file should not be created.
-                    if detector_field not in v.data:
-                        detector_field_exists = False
-                        break
+                        # Check if detector field does not exist. If not, then the file should not be created.
+                        if detector_field not in v.data:
+                            detector_field_exists = False
+                            break
 
-                    print()
-                    print(f"Collecting data from detector '{detector_name}' (field '{detector_field}')")
+                        print()
+                        print(f"Collecting data from detector '{detector_name}' (field '{detector_field}')")
 
-                    # Determine the number of channels from the size of the table with fluorescence data
-                    num_det = v.data[detector_field].shape[1]
+                        # Determine the number of channels from the size of the table with fluorescence data
+                        num_det = v.data[detector_field].shape[1]
 
-                    # Now allocate space for fluorescence data
-                    if create_each_det is False:
-                        new_data['det_sum'] = np.zeros(new_shape)
-                    else:
-                        for i in range(num_det):
-                            new_data[f'det{i + 1}'] = np.zeros(new_shape)
+                        # Now allocate space for fluorescence data
+                        if create_each_det is False:
+                            new_data['det_sum'] = np.zeros(new_shape, dtype=np.float32)
+                        else:
+                            for i in range(num_det):
+                                new_data[f'det{i + 1}'] = np.zeros(new_shape, dtype=np.float32)
 
-                    print(f"Number of the detector channels: {num_det}")
+                        print(f"Number of the detector channels: {num_det}")
 
-                if m < datashape[0]:   # scan is not finished
-                    if save_scaler is True:
-                        for n in scaler_list[:-1] + [xpos_name]:
-                            min_len = min(v.data[n].size, datashape[1])
-                            data[n][m, :min_len] = v.data[n][:min_len]
-                            # position data or i0 has shorter length than fluor data
-                            if min_len < datashape[1]:
-                                len_diff = datashape[1] - min_len
-                                # interpolation on scaler data
-                                interp_list = (v.data[n][-1] - v.data[n][-3]) / 2 * \
-                                    np.arange(1, len_diff + 1) + v.data[n][-1]
-                                data[n][m, min_len:datashape[1]] = interp_list
-                    fluor_len = v.data[detector_field].shape[0]
-                    if m > 0 and not (m % 10):
-                        print(f"Processed {m} of {n_scan_lines_total} lines ...")
-                    # print(f"m = {m} Data shape {v.data['fluor'].shape} - {v.data['fluor'].shape[1] }")
-                    # print(f"Data keys: {v.data.keys()}")
-                    if create_each_det is False:
-                        for i in range(num_det):
-                            # in case the data length in each line is different
-                            new_data['det_sum'][m, :fluor_len, :] += v.data[detector_field][:, i, :]
-                    else:
-                        for i in range(num_det):
-                            # in case the data length in each line is different
-                            new_data['det'+str(i+1)][m, :fluor_len, :] = v.data[detector_field][:, i, :]
+                    if m < datashape[0]:   # scan is not finished
+                        if save_scaler is True:
+                            for n in scaler_list[:-1] + [xpos_name]:
+                                min_len = min(v.data[n].size, datashape[1])
+                                data[n][m, :min_len] = v.data[n][:min_len]
+                                # position data or i0 has shorter length than fluor data
+                                if min_len < datashape[1]:
+                                    len_diff = datashape[1] - min_len
+                                    # interpolation on scaler data
+                                    interp_list = (v.data[n][-1] - v.data[n][-3]) / 2 * \
+                                        np.arange(1, len_diff + 1) + v.data[n][-1]
+                                    data[n][m, min_len:datashape[1]] = interp_list
+                        fluor_len = v.data[detector_field].shape[0]
+                        if m > 0 and not (m % 10):
+                            print(f"Processed {m} of {n_scan_lines_total} lines ...")
+                        # print(f"m = {m} Data shape {v.data['fluor'].shape} - {v.data['fluor'].shape[1] }")
+                        # print(f"Data keys: {v.data.keys()}")
+                        if create_each_det is False:
+                            for i in range(num_det):
+                                # in case the data length in each line is different
+                                new_data['det_sum'][m, :fluor_len, :] += v.data[detector_field][:, i, :]
+                        else:
+                            for i in range(num_det):
+                                # in case the data length in each line is different
+                                new_data['det'+str(i+1)][m, :fluor_len, :] = v.data[detector_field][:, i, :]
+
+            except Exception as ex:
+                logger.error(f"Error occurred while reading data: {ex}. Trying to retrieve available data ...")
 
             # If the detector field does not exist, then try the next one from the list
             if not detector_field_exists:
@@ -909,13 +1090,14 @@ def map_data2D_srx(runid, fpath,
             if output_to_file:
                 # output to file
                 print(f"Saving data to hdf file #{n_detectors_found}: Detector: {detector_name}.")
-                fpath_out = write_db_to_hdf_base(fpath_out, new_data,
+                fpath_out = write_db_to_hdf_base(fpath_out, new_data, metadata=mdata,
                                                  fname_add_version=fname_add_version,
                                                  file_overwrite_existing=file_overwrite_existing,
                                                  create_each_det=create_each_det)
 
             # Preparing data for the detector ``detector_name`` for output
-            d_dict = {"dataset": new_data, "file_name": fpath_out, "detector_name": detector_name}
+            d_dict = {"dataset": new_data, "file_name": fpath_out,
+                      "detector_name": detector_name, "metadata": mdata}
             data_output.append(d_dict)
 
         print()
@@ -988,7 +1170,17 @@ def map_data2D_tes(runid, fpath,
     """
 
     hdr = db[runid]
-    # start_doc = hdr['start']
+    runid = hdr.start["scan_id"]  # Replace with the true value (runid may be relative, such as -2)
+
+    # The dictionary holding scan metadata
+    mdata = _extract_metadata_from_header(hdr)
+    # Some metadata is located at specific places in the descriptor documents
+    # Search through the descriptor documents for the metadata
+    v = _get_metadata_from_descriptor_document(hdr, data_key="mono_energy", stream_name="baseline")
+    # Incident energy in the descriptor document is expected to be more accurate, so
+    #   overwrite the value if it already exists
+    if v is not None:
+        mdata["instrument_mono_incident_energy"] = v / 1000.0  # eV to keV
 
     if completed_scans_only and not _is_scan_complete(hdr):
         raise Exception("Scan is incomplete. Only completed scans are currently processed.")
@@ -1059,7 +1251,25 @@ def map_data2D_tes(runid, fpath,
             s_data = hdr.table()[name]
             # Convert pandas dataframe to a list of ndarrays (.to_numpy())
             #   and then stack the arrays into a single 2D array
-            s_data = np.vstack(s_data.to_numpy())
+            s_data = s_data.to_numpy()
+
+            # Fix for the issue: 'empty' rows in scaler data. Fill 'empty' row
+            #   with the nearest (preceding) row.
+            # TODO: investigate the issue of 'empty' scaler ('dwell_time') rows at TES
+            n_full = -1
+            for _n in range(len(s_data)):
+                if s_data[_n].shape != ():
+                    n_full = _n
+                    break
+            for _n in range(len(s_data)):
+                if s_data[_n].shape == ():
+                    s_data[_n] = np.copy(s_data[n_full])
+                    logger.error(f"Scaler '{name}': row #{_n} contains no data. "
+                                 f"Replaced by data from row #{n_full}")
+                else:
+                    n_full = _n
+
+            s_data = np.vstack(s_data)
             if scaler_data is None:
                 data_shape = s_data.shape
                 scaler_data = np.zeros(shape=data_shape + (n_scalers,), dtype=float)
@@ -1077,7 +1287,7 @@ def map_data2D_tes(runid, fpath,
     detector_field = "fluor"
 
     # Read detector values (for single detector)
-    detector_data = np.zeros(shape=data_shape + (spectrum_len,), dtype=float)
+    detector_data = np.zeros(shape=data_shape + (spectrum_len,), dtype=np.float32)
     n_events = data_shape[0]
     n_events_found = 0
     e = hdr.events(fill=True, stream_name="primary")
@@ -1086,12 +1296,18 @@ def map_data2D_tes(runid, fpath,
             print("The number of lines is less than expected")
             break
         data = v.data[detector_field]
-        data_det1 = np.array(data[:, 0, :])
+        data_det1 = np.array(data[:, 0, :], dtype=np.float32)
         detector_data[n, :, :] = data_det1
         n_events_found = n + 1
     if n_events_found < n_events:
         print("The number of lines is less than expected. The experiment may be incomplete")
-    new_data['det1'] = detector_data
+    # Note: the following code assumes that the detector has only one channel.
+    #   If the detector is upgraded, the following code will have to be rewritten, but
+    #   the rest of the data loading procedure will have to be modified anyway.
+    if create_each_det:
+        new_data['det1'] = detector_data
+    else:
+        new_data['det_sum'] = detector_data
 
     num_det = 1
     detector_name = "xs"
@@ -1108,12 +1324,13 @@ def map_data2D_tes(runid, fpath,
     if output_to_file:
         # output to file
         print(f"Saving data to hdf file #{n_detectors_found}: Detector: {detector_name}.")
-        write_db_to_hdf_base(fpath_out, new_data,
+        write_db_to_hdf_base(fpath_out, new_data, metadata=mdata,
                              fname_add_version=fname_add_version,
                              file_overwrite_existing=file_overwrite_existing,
                              create_each_det=create_each_det)
 
-    d_dict = {"dataset": new_data, "file_name": fpath_out, "detector_name": detector_name}
+    d_dict = {"dataset": new_data, "file_name": fpath_out,
+              "detector_name": detector_name, "metadata": mdata}
     data_output.append(d_dict)
 
     return data_output
@@ -1171,15 +1388,22 @@ def map_data2D_xfm(runid, fpath,
     dict of data in 2D format matching x,y scanning positions
     """
     hdr = db[runid]
+    runid = hdr.start["scan_id"]  # Replace with the true value (runid may be relative, such as -2)
 
     if completed_scans_only and not _is_scan_complete(hdr):
         raise Exception("Scan is incomplete. Only completed scans are currently processed.")
 
+    # Generate the default file name for the scan
+    if fpath is None:
+        fpath = f"scan2D_{runid}.h5"
+
     # Output data is the list of data structures for all available detectors
     data_output = []
 
     # spectrum_len = 4096
     start_doc = hdr['start']
+    # The dictionary holding scan metadata
+    mdata = _extract_metadata_from_header(hdr)
     plan_n = start_doc.get('plan_name')
     if 'fly' not in plan_n:  # not fly scan
         datashape = start_doc['shape']   # vertical first then horizontal
@@ -1203,17 +1427,19 @@ def map_data2D_xfm(runid, fpath,
                               datashape,
                               det_list=xrf_detector_names,
                               pos_list=hdr.start.motors,
+                              create_each_det=create_each_det,
                               scaler_list=config_data['scaler_list'],
                               fly_type=fly_type)
         if output_to_file:
             print('Saving data to hdf file.')
-            write_db_to_hdf_base(fpath, data_out,
+            write_db_to_hdf_base(fpath, data_out, metadata=mdata,
                                  fname_add_version=fname_add_version,
                                  file_overwrite_existing=file_overwrite_existing,
                                  create_each_det=create_each_det)
 
         detector_name = "xs"
-        d_dict = {"dataset": data_out, "file_name": fpath, "detector_name": detector_name}
+        d_dict = {"dataset": data_out, "file_name": fpath,
+                  "detector_name": detector_name, "metadata": mdata}
         data_output.append(d_dict)
 
         return data_output
@@ -1403,6 +1629,7 @@ def assemble_data_SRX_stepscan(
             new_v_shape = len(channel_data) // datashape[1]
 
             new_data = np.vstack(channel_data)
+            new_data = new_data.astype(np.float32, copy=False)  # Change representation to np.float32
             new_data = new_data[:new_v_shape*datashape[1], :]
 
             new_data = new_data.reshape([new_v_shape, datashape[1],
@@ -1495,6 +1722,7 @@ def map_data2D(data, datashape,
                det_list=('xspress3_ch1', 'xspress3_ch2', 'xspress3_ch3'),
                pos_list=('zpssx[um]', 'zpssy[um]'),
                scaler_list=('sclr1_ch3', 'sclr1_ch4'),
+               create_each_det=False,
                fly_type=None, subscan_dims=None, spectrum_len=4096):
     """
     Data is obained from databroker. Transfer items from data to a dictionay of
@@ -1539,18 +1767,21 @@ def map_data2D(data, datashape,
             # new veritcal shape is defined to ignore zeros points caused by stopped/aborted scans
             new_v_shape = len(channel_data) // datashape[1]
             new_data = np.vstack(channel_data)
+            new_data = new_data.astype(np.float32, copy=False)  # Change representation to np.float32
             new_data = new_data[:new_v_shape*datashape[1], :]
             new_data = new_data.reshape([new_v_shape, datashape[1],
                                          len(channel_data[1])])
             if new_data.shape[2] != spectrum_len:
                 # merlin detector has spectrum len 2048
                 # make all the spectrum len to 4096, to avoid unpredicted error in fitting part
-                new_tmp = np.zeros([new_data.shape[0], new_data.shape[1], spectrum_len])
+                new_tmp = np.zeros([new_data.shape[0], new_data.shape[1], spectrum_len],
+                                   dtype=np.float32)
                 new_tmp[:, :, :new_data.shape[2]] = new_data
                 new_data = new_tmp
             if fly_type in ('pyramid',):
                 new_data = flip_data(new_data, subscan_dims=subscan_dims)
-            data_output[detname] = new_data
+            if create_each_det:
+                data_output[detname] = new_data
             if sum_data is None:
                 # Note: Here is the place where the error was found!!!
                 #   The assignment in the next line used to be written as
@@ -1620,7 +1851,7 @@ def _get_fpath_not_existing(fpath):
     return fpath
 
 
-def write_db_to_hdf_base(fpath, data,
+def write_db_to_hdf_base(fpath, data, *, metadata=None,
                          fname_add_version=False,
                          file_overwrite_existing=False,
                          create_each_det=True):
@@ -1651,11 +1882,29 @@ def write_db_to_hdf_base(fpath, data,
       key 'pos_names' - the list of position field names, must have entries 'x_pos' and 'y_pos'
       key 'pos_data' - 3D ndarray, 1st index matches the position in 'pos_names' list
     """
+
     interpath = 'xrfmap'
     sum_data = None
     xrf_det_list = [n for n in data.keys() if 'det' in n and 'sum' not in n]
     xrf_det_list.sort()
 
+    # Verify that raw fluorescence data is represented with np.float32 precision: print the warning message
+    #   and convert the raw spectrum data to np.float32. Assume that data is represented as ndarray.
+    def incorrect_type_msg(channel, data_type):
+        logger.warning(f"Attemptying to save raw fluorescence data for the channel '{channel}' "
+                       f"as '{data_type}' numbers.\n    Memory may be used inefficiently. "
+                       f"Please, inform the PyXRF developers.\n    The data is converted from '{data_type}' "
+                       f"to 'np.float32' before saving to file.")
+    if "det_sum" in data and isinstance(data["det_sum"], np.ndarray):
+        if data["det_sum"].dtype != np.float32:
+            incorrect_type_msg("det_sum", data["det_sum"].dtype)
+            data["det_sum"] = data["det_sum"].astype(np.float32, copy=False)
+    for detname in xrf_det_list:
+        if detname in data and isinstance(data[detname], np.ndarray):
+            if data[detname].dtype != np.float32:
+                incorrect_type_msg(detname, data[detname].dtype)
+                data[detname] = data[detname].astype(np.float32, copy=False)
+
     file_open_mode = "a"
     if os.path.exists(fpath):
         if fname_add_version:
@@ -1670,6 +1919,26 @@ def write_db_to_hdf_base(fpath, data,
                 raise IOError(f"'write_db_to_hdf_base': File '{fpath}' already exists.")
 
     with h5py.File(fpath, file_open_mode) as f:
+
+        # Create metadata group
+        metadata_grp = f.create_group(f"{interpath}/scan_metadata")
+        # This group of attributes are always created. It doesn't matter if metadata
+        #   is provided to the function.
+        metadata_grp.attrs["file_type"] = "XRF-MAP"
+        metadata_grp.attrs["file_format"] = "NSLS2-XRF-MAP"
+        metadata_grp.attrs["file_format_version"] = "1.0"
+        metadata_grp.attrs["file_software"] = "PyXRF"
+        metadata_grp.attrs["file_software_version"] = pyxrf_version
+        # Present time in NEXUS format (should it be UTC time)?
+        metadata_grp.attrs["file_created_time"] = ttime.strftime("%Y-%m-%dT%H:%M:%S+00:00", ttime.localtime())
+
+        # Now save the rest of the scan metadata if metadata is provided
+        if metadata:
+            # We assume, that metadata does not contain repeated keys. Otherwise the
+            #   entry with the last occurrence of the key will override the previous ones.
+            for key, value in metadata.items():
+                metadata_grp.attrs[key] = value
+
         if create_each_det is True:
             for detname in xrf_det_list:
                 new_data = data[detname]
